theory CoSP_PKES_Computational imports
  CoSP_PKES_Symbolic
  Length_Regularity
  IND_CCA2_SUF_CMA
begin

abbreviation sample_bitstring :: "nat \<Rightarrow> bitstring pmf"
where "sample_bitstring n \<equiv> pmf_of_set (nlists UNIV n)"

record tagging =
  \<T>_nonce :: "security \<Rightarrow> bitstring \<Rightarrow> bool"
  \<T>_cipher :: "security \<Rightarrow> bitstring \<Rightarrow> bool"
  \<T>_ekey :: "security \<Rightarrow> bitstring \<Rightarrow> bool"
  \<T>_dkey :: "security \<Rightarrow> bitstring \<Rightarrow> bool"
  \<T>_signature :: "security \<Rightarrow> bitstring \<Rightarrow> bool"
  \<T>_vkey :: "security \<Rightarrow> bitstring \<Rightarrow> bool"
  \<T>_sigkey :: "security \<Rightarrow> bitstring \<Rightarrow> bool"
  \<T>_pair :: "security \<Rightarrow> bitstring \<Rightarrow> bool"
  \<T>_string :: "security \<Rightarrow> bitstring \<Rightarrow> bool"
  -- \<open>All nonces have same length for any given security parameter (impl. cond. 3)
    @{const size_nonce} returns the number of random bits in a nonce rather than the length
    of the encoded bitstring.
    \<close>
  size_nonce :: "security \<Rightarrow> nat"

subsection {* Computational implementation *}

type_synonym ekey = bitstring
type_synonym dkey = bitstring
type_synonym sigkey = bitstring
type_synonym vkey = bitstring
type_synonym plain = bitstring
type_synonym cipher = bitstring
type_synonym message = bitstring
type_synonym signature = bitstring

text \<open>
  There are two takes on randomness: extrinsic and intrinsic.
  In the extrinsic approach, functions are deterministic, but take an additional parameter as
  the source of randomness, i.e., they are like random variables. In the intrinsic view,
  the functions themselves are distributions.

  In the definitions of security notions and most of HOL-Probability, we follow the intrinsic view.
  CoSP (and most computational soundness results) take the extrinsic view. We can transform the
  extrinsic view into the intrinsic one by mapping the random variable over the distribution of
  input values. The other direction is not as easy because we have to ensure that the input 
  parameter contains enough randomness.
\<close>

type_synonym rand = bitstring

record implementation = tagging +
  -- \<open>public key encryption scheme\<close>
  A_ek :: "security \<Rightarrow> rand \<Rightarrow> ekey option"
  A_dk :: "security \<Rightarrow> rand \<Rightarrow> dkey option"
  A_e :: "security \<Rightarrow> ekey \<Rightarrow> bitstring \<Rightarrow> rand \<Rightarrow> cipher option"
  A_d :: "security \<Rightarrow> dkey \<Rightarrow> cipher \<Rightarrow> bitstring option"
  A_ekof :: "security \<Rightarrow> cipher \<Rightarrow> ekey option"
  -- \<open>signature scheme\<close>
  A_vk :: "security \<Rightarrow> rand \<Rightarrow> vkey option"
  A_sk :: "security \<Rightarrow> rand \<Rightarrow> sigkey option"
  A_sig :: "security \<Rightarrow> sigkey \<Rightarrow> bitstring \<Rightarrow> rand \<Rightarrow> signature option"
  A_verify :: "security \<Rightarrow> vkey \<Rightarrow> signature \<Rightarrow> bitstring option"
  A_vkof :: "security \<Rightarrow> signature \<Rightarrow> vkey option"
  -- \<open>tag as a nonce (given as a random bitstring)\<close>
  A_nonce :: "security \<Rightarrow> bitstring \<Rightarrow> bitstring"
  -- \<open>tuples\<close>
  A_pair :: "security \<Rightarrow> bitstring \<Rightarrow> bitstring \<Rightarrow> bitstring" -- \<open>impl. cond. 6 requires that @{term A_pair} is total\<close>
  A_fst :: "security \<Rightarrow> bitstring \<Rightarrow> bitstring option"
  A_snd :: "security \<Rightarrow> bitstring \<Rightarrow> bitstring option"
  -- \<open>payload strings\<close>
  A_empty :: "security \<Rightarrow> bitstring option"
  A_string :: "security \<Rightarrow> bool \<Rightarrow> bitstring \<Rightarrow> bitstring option"
  A_unstring :: "security \<Rightarrow> bool \<Rightarrow> bitstring \<Rightarrow> bitstring option"
  -- \<open>garbage\<close>
  A_garbage :: "security \<Rightarrow> bitstring \<Rightarrow> bitstring option"
  A_garbageE :: "security \<Rightarrow> bitstring \<Rightarrow> bitstring \<Rightarrow> bitstring option"
  A_garbageSIG :: "security \<Rightarrow> bitstring \<Rightarrow> bitstring \<Rightarrow> bitstring option"

definition A_isenc :: "'a implementation_scheme \<Rightarrow> security \<Rightarrow> bitstring \<Rightarrow> cipher option"
where "A_isenc A \<eta> m = (if \<T>_cipher A \<eta> m then Some m else None)"

definition A_issig :: "'a implementation_scheme \<Rightarrow> security \<Rightarrow> bitstring \<Rightarrow> ekey option"
where "A_issig A \<eta> m = (if \<T>_signature A \<eta> m then Some m else None)"

definition A_isek :: "'a implementation_scheme \<Rightarrow> security \<Rightarrow> bitstring \<Rightarrow> ekey option"
where "A_isek A \<eta> m = (if \<T>_ekey A \<eta> m then Some m else None)"

definition A_isvk :: "'a implementation_scheme \<Rightarrow> security \<Rightarrow> bitstring \<Rightarrow> vkey option"
where "A_isvk A \<eta> m = (if \<T>_vkey A \<eta> m then Some m else None)"

definition A_equals :: "'a implementation_scheme \<Rightarrow> security \<Rightarrow> bitstring \<Rightarrow> bitstring \<Rightarrow> bitstring option"
where "A_equals A \<eta> m1 m2 = (if m1 = m2 then Some m1 else None)"

locale comp_impl_base =
  fixes A :: implementation

locale comp_impl = comp_impl_base A
  for A :: implementation
  +
  -- \<open>Disjoint types (impl. cond. 2)\<close>
  assumes \<T>_disjoint: 
    "disjointp [\<T>_nonce A \<eta>, \<T>_cipher A \<eta>, \<T>_ekey A \<eta>, \<T>_dkey A \<eta>, \<T>_signature A \<eta>, \<T>_vkey A \<eta>,
                \<T>_sigkey A \<eta>, \<T>_pair A \<eta>, \<T>_string A \<eta>]"

  -- \<open>length regularity (impl. cond. 3\<close>
  assumes length_reg_A_e: "length_regular (A_e A \<eta>)"
  and length_reg_A_ek: "length_regular (A_ek A \<eta>)"
  and length_reg_A_dk: "length_regular (A_dk A \<eta>)"
  and length_reg_A_sig: "length_regular (A_sig A \<eta>)"
  and length_reg_A_vk: "length_regular (A_vk A \<eta>)"
  and length_reg_A_sk: "length_regular (A_sk A \<eta>)"
  and length_reg_A_pair: "length_regular (A_pair A \<eta>)"
  and length_reg_A_string: "length_regular (A_string A \<eta>)"
  assumes length_A_nonce:
  "\<lbrakk> length xs = size_nonce A \<eta>; length ys = size_nonce A \<eta> \<rbrakk> 
  \<Longrightarrow> length (A_nonce A \<eta> xs) = length (A_nonce A \<eta> ys)"

  -- \<open>type correctness (impl. cond. 5)\<close>
  assumes \<T>_A_ek: "ekey \<in> set_option (A_ek A \<eta> rnd) \<Longrightarrow> \<T>_ekey A \<eta> ekey"
  and \<T>_A_dk: "dkey \<in> set_option (A_dk A \<eta> rnd) \<Longrightarrow> \<T>_dkey A \<eta> dkey"
  and \<T>_A_e: "cipher \<in> set_option (A_e A \<eta> ekey m rnd) \<Longrightarrow> \<T>_cipher A \<eta> cipher"
  and \<T>_A_vk: "vkey \<in> set_option (A_vk A \<eta> rnd) \<Longrightarrow> \<T>_vkey A \<eta> vkey"
  and \<T>_A_sk: "sigkey \<in> set_option (A_sk A \<eta> sigkey) \<Longrightarrow> \<T>_sigkey A \<eta> sigkey"
  and \<T>_A_sig: "sig \<in> set_option (A_sig A \<eta> sigkey m rnd) \<Longrightarrow> \<T>_signature A \<eta> sig"
  and \<T>_A_ekof: "ekey \<in> set_option (A_ekof A \<eta> m) \<Longrightarrow> \<T>_ekey A \<eta> ekey"
  and \<T>_A_vkof: "vk \<in> set_option (A_vkof A \<eta> signature) \<Longrightarrow> \<T>_vkey A \<eta> vk"
  and \<T>_A_nonce: "length xs = size_nonce A \<eta> \<Longrightarrow> \<T>_nonce A \<eta> (A_nonce A \<eta> xs)"
  and \<T>_A_empty: "m \<in> set_option (A_empty A \<eta>) \<Longrightarrow> \<T>_string A \<eta> m"
  and \<T>_A_string: "string \<in> set_option (A_string A \<eta> b m) \<Longrightarrow> \<T>_string A \<eta> string"

  -- \<open>pairs (impl. cond. 6)\<close>
  assumes A_fst_pair: "A_fst A \<eta> (A_pair A \<eta> m1 m2) = Some m1"
  and A_snd_pair: "A_snd A \<eta> (A_pair A \<eta> m1 m2) = Some m2"
  and A_pair_exhaust: "\<T>_pair A \<eta> m \<Longrightarrow> \<exists>m1 m2. m = A_pair A \<eta> m1 m2"
  and A_fst_fail: "\<not> \<T>_pair A \<eta> m \<Longrightarrow> A_fst A \<eta> m = None"
  and A_snd_fail: "\<not> \<T>_pair A \<eta> m \<Longrightarrow> A_snd A \<eta> m = None"

  -- \<open>strings (impl. cond. 7)\<close>
  assumes A_unstring_string: "\<T>_string A \<eta> m \<Longrightarrow> A_string A \<eta> b m \<bind> A_unstring A \<eta> b = Some m"
  and A_unstring_string_fail: "\<lbrakk> \<T>_string A \<eta> m; b \<noteq> b' \<rbrakk> \<Longrightarrow> A_string A \<eta> b m \<bind> A_unstring A \<eta> b' = None"
  and A_unstring_empty: "A_empty A \<eta> = Some m \<Longrightarrow> A_unstring A \<eta> b m = None"
  and A_unstring_fail: "\<not> \<T>_string A \<eta> m \<Longrightarrow> A_unstring A \<eta> b m = None"

  -- \<open>encryption key extraction (impl. cond. 8)\<close>
  assumes A_ekof_E: "\<lbrakk> \<T>_ekey A \<eta> p; \<T>_nonce A \<eta> y \<rbrakk> \<Longrightarrow> A_e A \<eta> p m y \<bind> A_ekof A \<eta> = Some p"
  and A_ekof_eq_None_iff: "A_ekof A \<eta> m = None \<longleftrightarrow> \<not> \<T>_cipher A \<eta> m"

  -- \<open>new implementation condition: needed for the proof of indistinguishability of sim1 and sim1'
    because sim1' uses the oracle for generating encryption keys, and the oracle always generates
    the appropriate decryption key. So we must ensure that we can generate the key pair whenever we
    can generate one component. Similarly for signature keys.\<close>
  and A_ek_dk_None: "A_ek A \<eta> n = None \<longleftrightarrow> A_dk A \<eta> n = None"
  and A_sk_vk_None: "A_sk A \<eta> n = None \<longleftrightarrow> A_vk A \<eta> n = None"

  -- \<open>verification key extraction (impl. cond. 9)\<close>
  assumes A_vkof_SIG: "\<lbrakk> \<T>_nonce A \<eta> x; \<T>_nonce A \<eta> z \<rbrakk> \<Longrightarrow> A_sk A \<eta> x \<bind> (\<lambda>x. A_sig A \<eta> x y z \<bind> A_vkof A \<eta>) = A_vk A \<eta> x"
  and A_vkof_eq_None_iff: "A_vkof A \<eta> m = None \<longleftrightarrow> \<not> \<T>_signature A \<eta> m"

  -- \<open>encryption failure (impl. cond. 10)\<close>
  assumes A_e_eq_NoneI: "\<not> \<T>_ekey A \<eta> p \<Longrightarrow> A_e A \<eta> p m y = None"
  
  -- \<open>decryption failure (impl. cond. 11)\<close>
  assumes A_d_eq_NoneI: "\<lbrakk> A_dk A \<eta> r = Some dkey; \<T>_nonce A \<eta> r; A_ekof A \<eta> m \<noteq> A_ek A \<eta> r \<rbrakk> \<Longrightarrow> A_d A \<eta> dkey m = None"

  -- \<open>decryption correct (impl. cond. 12)\<close>
  assumes A_d_e:
  "\<lbrakk> \<T>_nonce A \<eta> r; \<T>_nonce A \<eta> r' \<rbrakk>
  \<Longrightarrow> A_dk A \<eta> r \<bind> (\<lambda>dkey. A_ek A \<eta> r \<bind> (\<lambda>ekey. A_e A \<eta> ekey m r' \<bind> A_d A \<eta> dkey)) = Some m"
  
  -- \<open>signature verification (impl. cond. 13)\<close>
  assumes A_verify_sig:
  "\<lbrakk> \<T>_nonce A \<eta> r; \<T>_nonce A \<eta> r' \<rbrakk>
  \<Longrightarrow> A_vk A \<eta> r \<bind> (\<lambda>vkey. A_sk A \<eta> r \<bind> (\<lambda>sigkey. A_sig A \<eta> skey m r' \<bind> A_verify A \<eta> vkey)) = Some m"

  -- \<open>verification key extraction (impl. cond. 14)\<close>
  assumes A_vkof_verify: "A_verify A \<eta> p s \<noteq> None \<Longrightarrow> A_vkof A \<eta> s = Some p"

  -- \<open>test on encryption keys (impl. cond. 15)\<close>
  assumes A_isek_eq_SomeI: "\<T>_ekey A \<eta> m \<Longrightarrow> A_isek A \<eta> m = Some m"
  and A_isek_eq_NoneI: "\<not> \<T>_ekey A \<eta> m \<Longrightarrow> A_isek A \<eta> m = None"

  -- \<open>test on verification keys (impl. cond. 16)\<close>
  assumes A_isvk_eq_SomeI: "\<T>_vkey A \<eta> m \<Longrightarrow> A_isvk A \<eta> m = Some m"
  and A_isvk_eq_NoneI: "\<not> \<T>_vkey A \<eta> m \<Longrightarrow> A_isvk A \<eta> m = None"

  -- \<open>test on ciphertexts (impl. cond. 17)\<close>
  assumes A_isenc_eq_SomeI: "\<T>_cipher A \<eta> m \<Longrightarrow> A_isenc A \<eta> m = Some m"
  and A_isenc_eq_NoneI: "\<not> \<T>_cipher A \<eta> m \<Longrightarrow> A_isenc A \<eta> m = None"
  
  -- \<open>test on signatures (impl. cond. 18)\<close>
  assumes A_issig_eq_SomeI: "\<T>_signature A \<eta> m \<Longrightarrow> A_issig A \<eta> m = Some m"
  and A_issig_eq_NoneI: "\<not> \<T>_signature A \<eta> m \<Longrightarrow> A_issig A \<eta> m = None"

  -- \<open>encryption clash (impl. cond. 21)\<close>
  assumes A_e_clash:
  -- \<open>FIXME @{text ekey} should vary with @{text \<eta>}, but considering all functions into
       encryption keys does not seem sensible either\<close>
  "\<T>_ekey A \<eta> ekey \<Longrightarrow> negligible (\<lambda>\<eta>. 
     pmf (map_pmf (\<lambda>(r, r'). case (A_e A \<eta> ekey m (A_nonce A \<eta> r), A_e A \<eta> ekey m (A_nonce A \<eta> r')) of (Some cipher1, Some cipher2) \<Rightarrow> cipher1 = cipher2 | _ \<Rightarrow> False)
     (pair_pmf (sample_bitstring (size_nonce A \<eta>)) (sample_bitstring (size_nonce A \<eta>)))) True)"

  -- \<open>signature clash (impl. cond. 22)\<close>
  assumes A_sig_clash:
  -- \<open>FIXME @{text sigkey} should vary with @{text \<eta>}, but considering all functions into
       signature keys does not seem sensible either\<close>
  "\<lbrakk> \<T>_sigkey A \<eta> sigkey; A_sk A \<eta> r = Some sigkey \<rbrakk> \<Longrightarrow> negligible (\<lambda>\<eta>.
     pmf (map_pmf (\<lambda>(r, r'). case (A_sig A \<eta> sigkey m (A_nonce A \<eta> r), A_sig A \<eta> sigkey m (A_nonce A \<eta> r')) of (Some sig1, Some sig2) \<Rightarrow> sig1 = sig2 | _ \<Rightarrow> False)
     (pair_pmf (sample_bitstring (size_nonce A \<eta>)) (sample_bitstring (size_nonce A \<eta>)))) True)"
begin

text {*
  Link with the intrinsic-style security definitions IND_CCA2 and SUF_CMA:
  As keys and randomness in well-formed terms must be nonces, we can sample them
  afresh, because we know their length.
*}

definition key_gen_pke :: "security \<Rightarrow> (ekey \<times> dkey) spmf"
where
  "key_gen_pke \<eta> =
   map_pmf (\<lambda>rnd. let rnd' = A_nonce A \<eta> rnd in A_ek A \<eta> rnd' \<bind> (\<lambda>ekey. A_dk A \<eta> rnd' \<bind> (\<lambda>dkey. Some (ekey, dkey))))
     (sample_bitstring (size_nonce A \<eta>))"

definition encrypt :: "security \<Rightarrow> ekey \<Rightarrow> plain \<Rightarrow> cipher spmf"
where "encrypt \<eta> ekey plain = map_pmf (A_e A \<eta> ekey plain \<circ> A_nonce A \<eta>) (sample_bitstring (size_nonce A \<eta>))"

abbreviation (input) decrypt :: "security \<Rightarrow> dkey \<Rightarrow> cipher \<Rightarrow> plain option"
where "decrypt \<equiv> A_d A"

abbreviation (input) valid_plain :: "security \<Rightarrow> plain \<Rightarrow> bool"
where "valid_plain \<eta> plain \<equiv> True" -- \<open>There does not seem to be a notion of message formats for plaintexts. We could take the domain of @{const A_e} here.\<close>

sublocale ind_cca2: ind_cca2 key_gen_pke encrypt decrypt valid_plain . -- \<open>Impl. cond. 19\<close>

definition key_gen_sig :: "security \<Rightarrow> (vkey \<times> sigkey) spmf"
where
  "key_gen_sig \<eta> = 
   map_pmf (\<lambda>rnd. let rnd' = A_nonce A \<eta> rnd in A_vk A \<eta> rnd' \<bind> (\<lambda>vkey. A_sk A \<eta> rnd' \<bind> (\<lambda>sigkey. Some (vkey, sigkey))))
     (sample_bitstring (size_nonce A \<eta>))"

definition sign :: "security \<Rightarrow> sigkey \<Rightarrow> message \<Rightarrow> signature spmf"
where "sign \<eta> sigkey m = map_pmf (\<lambda>r. A_sig A \<eta> sigkey m (A_nonce A \<eta> r)) (sample_bitstring (size_nonce A \<eta>))"

definition verify :: "security \<Rightarrow> vkey \<Rightarrow> message \<Rightarrow> signature \<Rightarrow> bool"
where "verify \<eta> vkey m signature \<longleftrightarrow> (case A_verify A \<eta> vkey signature of Some m' \<Rightarrow> m = m' | None \<Rightarrow> False)"

abbreviation (input) valid_message :: "security \<Rightarrow> signature \<Rightarrow> bool"
where "valid_message \<eta> msg \<equiv> True" -- \<open>There does not seem to be a notion of message formats. We could take the domain of @{const A_sig} here\<close>

sublocale suf_cma: suf_cma key_gen_sig sign verify valid_message . -- \<open>Impl. cond. 20\<close>

sublocale ind_cca2_suf_cma key_gen_pke encrypt decrypt valid_plain key_gen_sig sign verify valid_message .

type_synonym call = "(protocol_nonce, plain, cipher, protocol_nonce, message) ind_cca2_suf_cma.call\<^sub>n"
type_synonym ret = "(ekey, plain, cipher, vkey, signature) ret"
type_synonym state_oracle = "(protocol_nonce, ekey, dkey, cipher, protocol_nonce, vkey, sigkey, message, signature) state_oracle\<^sub>n"
type_synonym state_oracle\<^sub>1 = "(ekey, dkey, cipher, vkey, sigkey, message, signature) state_oracle\<^sub>1"
end
translations
  (type) "comp_impl.call" \<leftharpoondown> (type) "nat \<times> (unit + bool list + bool list \<times> bool list) + nat \<times> (unit + bool list)"
  (type) "comp_impl.ret" \<leftharpoondown> (type) "(bool list + bool list option + bool list) + bool list + bool list"
  (type) "comp_impl.state_oracle" \<leftharpoondown> 
  (type) "(nat \<Rightarrow> (bool list \<times> bool list \<times> bool list list) option) \<times>
          (nat \<Rightarrow> (bool list \<times> bool list \<times> (bool list \<times> bool list) list) option)"
  (type) "comp_impl.state_oracle\<^sub>1" \<leftharpoondown> 
  (type) "(bool list \<times> bool list \<times> bool list list) option \<times>
          (bool list \<times> bool list \<times> (bool list \<times> bool list) list) option"
context comp_impl begin
typ call typ ret typ state_oracle typ state_oracle\<^sub>1

abbreviation oracle\<^sub>n' where "oracle\<^sub>n' == oracle\<^sub>n"

definition "oracle" :: "bool \<Rightarrow> (state_oracle, call, ret) oracle"
where "oracle b \<eta> = (oracle\<^sub>n b \<eta>, (\<lambda>_. None, \<lambda>_. None))"

text {* Interface with symbolic implementation *}

fun A_C :: "constructor \<Rightarrow> security \<Rightarrow> bitstring list \<Rightarrow> bitstring option"
where
  "A_C E_C \<eta> ts = (case ts of [t1, t2, t3] \<Rightarrow> A_e A \<eta> t1 t2 t3 | _ \<Rightarrow> None)"
| "A_C EK_C \<eta> ts = (case ts of [t] \<Rightarrow> A_ek A \<eta> t | _ \<Rightarrow> None)"
| "A_C DK_C \<eta> ts = (case ts of [t] \<Rightarrow> A_dk A \<eta> t | _ \<Rightarrow> None)"
| "A_C SIG_C \<eta> ts = (case ts of [t1, t2, t3] \<Rightarrow> A_sig A \<eta> t1 t2 t3 | _ \<Rightarrow> None)"
| "A_C VK_C \<eta> ts = (case ts of [t] \<Rightarrow> A_vk A \<eta> t | _ \<Rightarrow> None)"
| "A_C SK_C \<eta> ts = (case ts of [t] \<Rightarrow> A_sk A \<eta> t | _ \<Rightarrow> None)"
| "A_C PAIR_C \<eta> ts = (case ts of [t1, t2] \<Rightarrow> Some (A_pair A \<eta> t1 t2) | _ \<Rightarrow> None)"
| "A_C EMPTY_C \<eta> ts = (case ts of [] \<Rightarrow> A_empty A \<eta> | _ \<Rightarrow> None)"
| "A_C (STRING_C b) \<eta> ts = (case ts of [t] \<Rightarrow> A_string A \<eta> b t | _ \<Rightarrow> None)"
| "A_C garbage_C \<eta> ts = (case ts of [t] \<Rightarrow> A_garbage A \<eta> t | _ \<Rightarrow> None)"
| "A_C garbageE_C \<eta> ts = (case ts of [t1, t2] \<Rightarrow> A_garbageE A \<eta> t1 t2 | _ \<Rightarrow> None)"
| "A_C garbageSIG_C \<eta> ts = (case ts of [t1, t2] \<Rightarrow> A_garbageSIG A \<eta> t1 t2 | _ \<Rightarrow> None)"

definition A_N :: "nonce \<Rightarrow> security \<Rightarrow> bitstring pmf"
where "A_N nonce \<eta> = map_pmf (A_nonce A \<eta>) (sample_bitstring (size_nonce A \<eta>))"

fun A_D :: "destructor \<Rightarrow> security \<Rightarrow> bitstring list \<Rightarrow> bitstring option"
where
  "A_D D \<eta> ts = (case ts of [t1, t2] \<Rightarrow> A_d A \<eta> t1 t2 | _ \<Rightarrow> None)"
| "A_D isenc \<eta> ts = (case ts of [t] \<Rightarrow> A_isenc A \<eta> t | _ \<Rightarrow> None)"
| "A_D isek \<eta> ts = (case ts of [t] \<Rightarrow> A_isek A \<eta> t | _ \<Rightarrow> None)"
| "A_D ekof \<eta> ts = (case ts of [t] \<Rightarrow> A_ekof A \<eta> t | _ \<Rightarrow> None)"
| "A_D destructor.verify \<eta> ts = (case ts of [t1, t2] \<Rightarrow> A_verify A \<eta> t1 t2 | _ \<Rightarrow> None)"
| "A_D issig \<eta> ts = (case ts of [t] \<Rightarrow> A_issig A \<eta> t | _ \<Rightarrow> None)"
| "A_D isvk \<eta> ts = (case ts of [t] \<Rightarrow> A_isvk A \<eta> t | _ \<Rightarrow> None)"
| "A_D vkof \<eta> ts = (case ts of [t] \<Rightarrow> A_vkof A \<eta> t | _ \<Rightarrow> None)"
| "A_D FST \<eta> ts = (case ts of [t] \<Rightarrow> A_fst A \<eta> t | _ \<Rightarrow> None)"
| "A_D SND \<eta> ts = (case ts of [t] \<Rightarrow> A_snd A \<eta> t | _ \<Rightarrow> None)"
| "A_D (unstring b) \<eta> ts = (case ts of [t] \<Rightarrow> A_unstring A \<eta> b t | _ \<Rightarrow> None)"
| "A_D equals \<eta> ts = (case ts of [t1, t2] \<Rightarrow> A_equals A \<eta> t1 t2 | _ \<Rightarrow> None)"

sublocale computational_implementation \<C> N \<D> \<T> deduce A_C A_N A_D . -- \<open>impl. cond. 1\<close>

lemma \<T>_disjoint':
  shows \<T>_nonce_cipher: "\<lbrakk> \<T>_nonce A \<eta> m; \<T>_cipher A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_nonce_ekey: "\<lbrakk> \<T>_nonce A \<eta> m; \<T>_ekey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_nonce_dkey: "\<lbrakk> \<T>_nonce A \<eta> m; \<T>_dkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_nonce_signature: "\<lbrakk> \<T>_nonce A \<eta> m; \<T>_signature A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_nonce_vkey: "\<lbrakk> \<T>_nonce A \<eta> m; \<T>_vkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_nonce_sigkey: "\<lbrakk> \<T>_nonce A \<eta> m; \<T>_sigkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_nonce_pair: "\<lbrakk> \<T>_nonce A \<eta> m; \<T>_pair A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_nonce_string: "\<lbrakk> \<T>_nonce A \<eta> m; \<T>_string A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_cipher_nonce: "\<lbrakk> \<T>_cipher A \<eta> m; \<T>_nonce A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_cipher_ekey: "\<lbrakk> \<T>_cipher A \<eta> m; \<T>_ekey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_cipher_dkey: "\<lbrakk> \<T>_cipher A \<eta> m; \<T>_dkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_cipher_signature: "\<lbrakk> \<T>_cipher A \<eta> m; \<T>_signature A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_cipher_vkey: "\<lbrakk> \<T>_cipher A \<eta> m; \<T>_vkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_cipher_sigkey: "\<lbrakk> \<T>_cipher A \<eta> m; \<T>_sigkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_cipher_pair: "\<lbrakk> \<T>_cipher A \<eta> m; \<T>_pair A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_cipher_string: "\<lbrakk> \<T>_cipher A \<eta> m; \<T>_string A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_ekey_nonce: "\<lbrakk> \<T>_ekey A \<eta> m; \<T>_nonce A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_ekey_cipher: "\<lbrakk> \<T>_ekey A \<eta> m; \<T>_cipher A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_ekey_dkey: "\<lbrakk> \<T>_ekey A \<eta> m; \<T>_dkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_ekey_signature: "\<lbrakk> \<T>_ekey A \<eta> m; \<T>_signature A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_ekey_vkey: "\<lbrakk> \<T>_ekey A \<eta> m; \<T>_vkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_ekey_sigkey: "\<lbrakk> \<T>_ekey A \<eta> m; \<T>_sigkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_ekey_pair: "\<lbrakk> \<T>_ekey A \<eta> m; \<T>_pair A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_ekey_string: "\<lbrakk> \<T>_ekey A \<eta> m; \<T>_string A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_dkey_nonce: "\<lbrakk> \<T>_dkey A \<eta> m; \<T>_nonce A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_dkey_cipher: "\<lbrakk> \<T>_dkey A \<eta> m; \<T>_cipher A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_dkey_ekey: "\<lbrakk> \<T>_dkey A \<eta> m; \<T>_ekey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_dkey_signature: "\<lbrakk> \<T>_dkey A \<eta> m; \<T>_signature A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_dkey_vkey: "\<lbrakk> \<T>_dkey A \<eta> m; \<T>_vkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_dkey_sigkey: "\<lbrakk> \<T>_dkey A \<eta> m; \<T>_sigkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_dkey_pair: "\<lbrakk> \<T>_dkey A \<eta> m; \<T>_pair A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_dkey_string: "\<lbrakk> \<T>_dkey A \<eta> m; \<T>_string A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_signature_nonce: "\<lbrakk> \<T>_signature A \<eta> m; \<T>_nonce A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_signature_cipher: "\<lbrakk> \<T>_signature A \<eta> m; \<T>_cipher A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_signature_ekey: "\<lbrakk> \<T>_signature A \<eta> m; \<T>_ekey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_signature_dkey: "\<lbrakk> \<T>_signature A \<eta> m; \<T>_dkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_signature_vkey: "\<lbrakk> \<T>_signature A \<eta> m; \<T>_vkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_signature_sigkey: "\<lbrakk> \<T>_signature A \<eta> m; \<T>_sigkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_signature_pair: "\<lbrakk> \<T>_signature A \<eta> m; \<T>_pair A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_signature_string: "\<lbrakk> \<T>_signature A \<eta> m; \<T>_string A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_vkey_nonce: "\<lbrakk> \<T>_vkey A \<eta> m; \<T>_nonce A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_vkey_cipher: "\<lbrakk> \<T>_vkey A \<eta> m; \<T>_cipher A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_vkey_ekey: "\<lbrakk> \<T>_vkey A \<eta> m; \<T>_ekey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_vkey_dkey: "\<lbrakk> \<T>_vkey A \<eta> m; \<T>_dkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_vkey_signature: "\<lbrakk> \<T>_vkey A \<eta> m; \<T>_signature A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_vkey_sigkey: "\<lbrakk> \<T>_vkey A \<eta> m; \<T>_sigkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_vkey_pair: "\<lbrakk> \<T>_vkey A \<eta> m; \<T>_pair A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_vkey_string: "\<lbrakk> \<T>_vkey A \<eta> m; \<T>_string A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_sigkey_nonce: "\<lbrakk> \<T>_sigkey A \<eta> m; \<T>_nonce A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_sigkey_cipher: "\<lbrakk> \<T>_sigkey A \<eta> m; \<T>_cipher A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_sigkey_ekey: "\<lbrakk> \<T>_sigkey A \<eta> m; \<T>_ekey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_sigkey_dkey: "\<lbrakk> \<T>_sigkey A \<eta> m; \<T>_dkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_sigkey_signature: "\<lbrakk> \<T>_sigkey A \<eta> m; \<T>_signature A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_sigkey_vkey: "\<lbrakk> \<T>_sigkey A \<eta> m; \<T>_vkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_sigkey_pair: "\<lbrakk> \<T>_sigkey A \<eta> m; \<T>_pair A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_sigkey_string: "\<lbrakk> \<T>_sigkey A \<eta> m; \<T>_string A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_pair_nonce: "\<lbrakk> \<T>_pair A \<eta> m; \<T>_nonce A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_pair_cipher: "\<lbrakk> \<T>_pair A \<eta> m; \<T>_cipher A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_pair_ekey: "\<lbrakk> \<T>_pair A \<eta> m; \<T>_ekey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_pair_dkey: "\<lbrakk> \<T>_pair A \<eta> m; \<T>_dkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_pair_signature: "\<lbrakk> \<T>_pair A \<eta> m; \<T>_signature A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_pair_vkey: "\<lbrakk> \<T>_pair A \<eta> m; \<T>_vkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_pair_sigkey: "\<lbrakk> \<T>_pair A \<eta> m; \<T>_sigkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_pair_string: "\<lbrakk> \<T>_pair A \<eta> m; \<T>_string A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_string_nonce: "\<lbrakk> \<T>_string A \<eta> m; \<T>_nonce A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_string_cipher: "\<lbrakk> \<T>_string A \<eta> m; \<T>_cipher A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_string_ekey: "\<lbrakk> \<T>_string A \<eta> m; \<T>_ekey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_string_dkey: "\<lbrakk> \<T>_string A \<eta> m; \<T>_dkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_string_signature: "\<lbrakk> \<T>_string A \<eta> m; \<T>_signature A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_string_vkeyg: "\<lbrakk> \<T>_string A \<eta> m; \<T>_vkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_string_sigkey: "\<lbrakk> \<T>_string A \<eta> m; \<T>_sigkey A \<eta> m \<rbrakk> \<Longrightarrow> False"
  and \<T>_string_pair: "\<lbrakk> \<T>_string A \<eta> m; \<T>_pair A \<eta> m \<rbrakk> \<Longrightarrow> False"
by((drule (1) disjointpD'[OF \<T>_disjoint], (rule nth_Cons_0 nth_eq_tlI)+, simp, simp, simp); fail)+

lemma A_type_correct:
  shows \<T>_A_ek': "A_ek A \<eta> rnd = Some ekey \<Longrightarrow> \<T>_ekey A \<eta> ekey"
  and \<T>_A_dk': "A_dk A \<eta> rnd = Some dkey \<Longrightarrow> \<T>_dkey A \<eta> dkey"
  and \<T>_A_e': "A_e A \<eta> ekey m rnd = Some cipher \<Longrightarrow> \<T>_cipher A \<eta> cipher"
  and \<T>_A_vk': "A_vk A \<eta> rnd = Some vkey \<Longrightarrow> \<T>_vkey A \<eta> vkey"
  and \<T>_A_sk': "A_sk A \<eta> sigkey = Some sigkey \<Longrightarrow> \<T>_sigkey A \<eta> sigkey"
  and \<T>_A_sig': "A_sig A \<eta> sigkey m rnd = Some sig \<Longrightarrow> \<T>_signature A \<eta> sig"
  and \<T>_A_ekof': "A_ekof A \<eta> m = Some ekey \<Longrightarrow> \<T>_ekey A \<eta> ekey"
  and \<T>_A_vkof': "A_vkof A \<eta> signature = Some vk \<Longrightarrow> \<T>_vkey A \<eta> vk"
  and \<T>_A_empty': "A_empty A \<eta> = Some m \<Longrightarrow> \<T>_string A \<eta> m"
  and \<T>_A_string': "A_string A \<eta> b m = Some string \<Longrightarrow> \<T>_string A \<eta> string"
  and \<T>_A_unstring': "A_unstring A \<eta> b m = Some string \<Longrightarrow> \<T>_string A \<eta> m"
by(auto 4 3 intro: \<T>_A_ek \<T>_A_dk \<T>_A_e \<T>_A_vk \<T>_A_sk \<T>_A_sig \<T>_A_ekof \<T>_A_vkof \<T>_A_empty \<T>_A_string dest: A_unstring_fail[where b=b] intro: ccontr)

lemma \<T>_cipher_A_ekofD: "\<T>_cipher A \<eta> cipher \<Longrightarrow> A_ekof A \<eta> cipher \<noteq> None"
by(simp add: A_ekof_eq_None_iff)

lemma \<T>_signature_A_vkofD: "\<T>_signature A \<eta> signature \<Longrightarrow> A_vkof A \<eta> signature \<noteq> None"
by(simp add: A_vkof_eq_None_iff)

lemma \<T>_A_pair [simp]: "\<T>_pair A \<eta> (A_pair A \<eta> m1 m2)"
using A_fst_pair[of \<eta> m1 m2] A_fst_fail[of \<eta> "A_pair A \<eta> m1 m2"] by auto

lemma A_ek_SomeD_dk: "A_ek A \<eta> n = Some m \<Longrightarrow> A_dk A \<eta> n \<noteq> None"
by(simp add: A_ek_dk_None[symmetric])

lemma A_dk_SomeD_ek: "A_dk A \<eta> n = Some m \<Longrightarrow> A_ek A \<eta> n \<noteq> None"
by(simp add: A_ek_dk_None)

lemma A_sk_SomeD_vk: "A_sk A \<eta> n = Some m \<Longrightarrow> A_vk A \<eta> n \<noteq> None"
by(simp add: A_sk_vk_None[symmetric])

lemma A_vk_SomeD_sk: "A_vk A \<eta> n = Some m \<Longrightarrow> A_sk A \<eta> n \<noteq> None"
by(simp add: A_sk_vk_None)

end

datatype msg_type
  = nonceTM
  | cipherTM
  | ekeyTM
  | signatureTM
  | vkeyTM  
  | pairTM
  | stringTM
  | garbageTM

context comp_impl_base begin

fun node_type_of_m :: "security \<Rightarrow> bitstring \<Rightarrow> msg_type"
where
  "node_type_of_m \<eta> m =
  (if \<T>_nonce A \<eta> m then nonceTM 
   else if \<T>_cipher A \<eta> m then cipherTM
   else if \<T>_ekey A \<eta> m then ekeyTM 
   else if \<T>_signature A \<eta> m then signatureTM
   else if \<T>_vkey A \<eta> m then vkeyTM
   else if \<T>_pair A \<eta> m then pairTM
   else if \<T>_string A \<eta> m then stringTM
   else garbageTM)"

declare node_type_of_m.simps [simp del]

end

context comp_impl begin

lemma node_type_of_m_simps:
  "\<T>_nonce A \<eta> m \<Longrightarrow> node_type_of_m \<eta> m = nonceTM"
  "\<T>_cipher A \<eta> m \<Longrightarrow> node_type_of_m \<eta> m = cipherTM"
  "\<T>_ekey A \<eta> m \<Longrightarrow> node_type_of_m \<eta> m = ekeyTM"
  "\<T>_signature A \<eta> m \<Longrightarrow> node_type_of_m \<eta> m = signatureTM"
  "\<T>_vkey A \<eta> m \<Longrightarrow> node_type_of_m \<eta> m = vkeyTM"
  "\<T>_pair A \<eta> m \<Longrightarrow> node_type_of_m \<eta> m = pairTM"
  "\<T>_string A \<eta> m \<Longrightarrow> node_type_of_m \<eta> m = stringTM"
  "\<lbrakk> \<not> \<T>_nonce A \<eta> m; \<not> \<T>_cipher A \<eta> m; \<not> \<T>_ekey A \<eta> m; \<not> \<T>_signature A \<eta> m; \<not> \<T>_vkey A \<eta> m; \<not> \<T>_pair A \<eta> m; \<not> \<T>_string A \<eta> m \<rbrakk> \<Longrightarrow> node_type_of_m \<eta> m = garbageTM"
by(subst node_type_of_m.simps; auto dest: \<T>_disjoint'; fail)+

end

subsubsection \<open>Length function\<close>

context comp_impl begin

text \<open>
  Adversary nonces are used to store arbitrary bitstrings in terms and serialisation extracts them.
  Therefore, @{term size_nonce} can only be used to compute the length of protocol nonces. This
  somehow violates the intuition that nonces should be sampled randomly and tagged with
  @{term A_nonce}. It is yet unclear whether this will create problems later.
\<close>

context fixes \<eta> :: nat begin

definition fixed_nonce :: "bitstring"
where "fixed_nonce = A_nonce A \<eta> (replicate (size_nonce A \<eta>) False)"

abbreviation len_of_P_nonce :: "nat"
where "len_of_P_nonce \<equiv> length fixed_nonce"

definition len_of_A_e :: "nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> nat"
where
  "len_of_A_e l1 l2 l3 =
  (THE n. n \<in> length ` (\<Union>m1\<in>nlists UNIV l1. \<Union>m2\<in>nlists UNIV l2. \<Union>m3\<in>nlists UNIV l3. set_option (A_e A \<eta> m1 m2 m3)))"

lemma len_of_A_e:
  "\<lbrakk> A_e A \<eta> m1 m2 m3 = Some m; l3 = length m3 \<rbrakk> \<Longrightarrow> len_of_A_e (length m1) (length m2) l3 = length m"
unfolding len_of_A_e_def
apply(rule the_equality)
 apply(fastforce simp add: nlists_alt_def)
apply(clarsimp simp add: nlists_alt_def)
apply(drule (4) length_reg_A_e[THEN length_reg_list_funD, THEN length_reg_list_funD, THEN length_reg_list_funD, THEN length_reg_optionD, rotated -2])
apply simp
done

definition len_of_A_ek :: "nat"
where "len_of_A_ek = (THE n. n \<in> length ` (\<Union>m\<in>nlists UNIV len_of_P_nonce. set_option (A_ek A \<eta> m)))"

lemma len_of_A_ek: "\<lbrakk> A_ek A \<eta> r_n = Some m; length r_n = len_of_P_nonce \<rbrakk> \<Longrightarrow> len_of_A_ek = length m"
unfolding len_of_A_ek_def
apply(rule the_equality)
 apply(fastforce simp add: nlists_alt_def)
apply(clarsimp simp add: nlists_alt_def)
apply(drule (1) length_reg_A_ek[THEN length_reg_list_funD, THEN length_reg_optionD, rotated -2]; simp_all)
done

definition len_of_A_dk :: "nat"
where "len_of_A_dk = (THE n. n \<in> length ` (\<Union>m\<in>nlists UNIV len_of_P_nonce. set_option (A_dk A \<eta> m)))"

lemma len_of_A_dk: "\<lbrakk> A_dk A \<eta> r_n = Some m; length r_n = len_of_P_nonce \<rbrakk> \<Longrightarrow> len_of_A_dk = length m"
unfolding len_of_A_dk_def
apply(rule the_equality)
 apply(fastforce simp add: nlists_alt_def)
apply(clarsimp simp add: nlists_alt_def)
apply(drule (1) length_reg_A_dk[THEN length_reg_list_funD, THEN length_reg_optionD, rotated -2]; simp_all)
done

definition len_of_A_sig :: "nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> nat"
where
  "len_of_A_sig l1 l2 l3 =
  (THE n. n \<in> length ` (\<Union>m1\<in>nlists UNIV l1. \<Union>m2\<in>nlists UNIV l2. \<Union>m3\<in>nlists UNIV l3. set_option (A_sig A \<eta> m1 m2 m3)))"

lemma len_of_A_sig:
  "A_sig A \<eta> m1 m2 m3 = Some m \<Longrightarrow> len_of_A_sig (length m1) (length m2) (length m3) = length m"
unfolding len_of_A_sig_def
apply(rule the_equality)
 apply(fastforce simp add: nlists_alt_def)
apply(clarsimp simp add: nlists_alt_def)
apply(drule (4) length_reg_A_sig[THEN length_reg_list_funD, THEN length_reg_list_funD, THEN length_reg_list_funD, THEN length_reg_optionD, rotated -2])
apply simp
done

definition len_of_A_vk :: "nat"
where "len_of_A_vk = (THE n. n \<in> length ` (\<Union>m\<in>nlists UNIV len_of_P_nonce. set_option (A_vk A \<eta> m)))"

lemma len_of_A_vk: "\<lbrakk> A_vk A \<eta> r_n = Some m; length r_n = len_of_P_nonce \<rbrakk> \<Longrightarrow> len_of_A_vk = length m"
unfolding len_of_A_vk_def
apply(rule the_equality)
 apply(fastforce simp add: nlists_alt_def)
apply(clarsimp simp add: nlists_alt_def)
apply(drule (1) length_reg_A_vk[THEN length_reg_list_funD, THEN length_reg_optionD, rotated -2]; simp_all)
done

definition len_of_A_sk :: "nat"
where "len_of_A_sk = (THE n. n \<in> length ` (\<Union>m\<in>nlists UNIV len_of_P_nonce. set_option (A_sk A \<eta> m)))"

lemma len_of_A_sk: "\<lbrakk> A_sk A \<eta> r_n = Some m; length r_n = len_of_P_nonce \<rbrakk> \<Longrightarrow> len_of_A_sk = length m"
unfolding len_of_A_sk_def
apply(rule the_equality)
 apply(fastforce simp add: nlists_alt_def)
apply(clarsimp simp add: nlists_alt_def)
apply(drule (1) length_reg_A_sk[THEN length_reg_list_funD, THEN length_reg_optionD, rotated -2]; simp_all)
done

definition len_of_A_pair :: "nat \<Rightarrow> nat \<Rightarrow> nat"
where "len_of_A_pair l1 l2 = length (A_pair A \<eta> (replicate l1 False) (replicate l2 False))"

lemma len_of_A_pair: "A_pair A \<eta> m1 m2 = m \<Longrightarrow> len_of_A_pair (length m1) (length m2) = length m"
by(auto intro: length_reg_A_pair[THEN length_reg_list_funD, THEN length_reg_list_funD, THEN length_reg_listD] simp add: len_of_A_pair_def)

definition len_of_A_string :: "bool \<Rightarrow> nat \<Rightarrow> nat"
where
  "len_of_A_string b l =
  (THE n. n \<in> length ` (\<Union>m\<in>nlists UNIV l. set_option (A_string A \<eta> b m)))"

lemma len_of_A_string: "A_string A \<eta> b m1 = Some m \<Longrightarrow> len_of_A_string b (length m1) = length m"
unfolding len_of_A_string_def
apply(rule the_equality)
 apply(fastforce simp add: nlists_alt_def)
apply(clarsimp simp add: nlists_alt_def)
apply(drule (1) length_reg_A_string[THEN length_reg_boolD, THEN length_reg_list_funD, THEN length_reg_optionD, rotated -2]; simp_all)
done

lemmas len_of_A_ =
  len_of_A_e len_of_A_ek len_of_A_dk len_of_A_sig len_of_A_vk len_of_A_sk len_of_A_pair len_of_A_string

fun len_of :: "t \<Rightarrow> nat"
where
  "len_of t = (case t of
     N (PNonce n) \<Rightarrow> length fixed_nonce
   | N (ANonce m) \<Rightarrow> length m
   | E t1 t2 (N (PNonce n)) \<Rightarrow> len_of_A_e (len_of t1) (len_of t2) len_of_P_nonce
   | E t1 t2 (N (ANonce m)) \<Rightarrow> length m
   | EK (N (PNonce n)) \<Rightarrow> len_of_A_ek
   | EK (N (ANonce m)) \<Rightarrow> length m
   | DK (N (PNonce n)) \<Rightarrow> len_of_A_dk
   | SIG t1 t2 (N (PNonce n)) \<Rightarrow> len_of_A_sig (len_of t1) (len_of t2) len_of_P_nonce
   | SIG t1 t2 (N (ANonce m)) \<Rightarrow> length m
   | VK (N (PNonce n)) \<Rightarrow> len_of_A_vk
   | VK (N (ANonce m)) \<Rightarrow> length m
   | SK (N (PNonce n)) \<Rightarrow> len_of_A_sk
   | PAIR t1 t2 \<Rightarrow> len_of_A_pair (len_of t1) (len_of t2)
   | STRING b t \<Rightarrow> len_of_A_string b (len_of t)
   | EMPTY \<Rightarrow> length (the (A_empty A \<eta>))
   | garbage (N (ANonce m)) \<Rightarrow> length m
   | garbageE t1 (N (ANonce m)) \<Rightarrow> length m
   | garbageSIG t1 (N (ANonce m)) \<Rightarrow> length m
   )"

declare len_of.simps[simp del]

simps_of_case len_of_simps [simp]: len_of.simps (splits: t.split nonce.split)

end

end

end