theory CoSP_PKES_Key_Safety imports
  CoSP_PKES_Computational
begin

subsection {* Key-safe protocols *}

text \<open>
  As node identifiers need not be unique, we cannot look back at what has happened.
  Therefore, we keep track of the path to the node in the protocol tree.

  The conditions for key-safety are slightly stronger than in the CoSP paper, because we additionally
  require a typing for nonces (which have identifiers). The usage of a nonce must be the same over
  all the protocol (rather than just for individual paths in the protocool). Similarly, a randomness
  nonce may be used only at one depth in the tree (counting only computation nodes) rather than
  arbitrarily deep in different paths. These restrictions are not significant, as one can always
  defer the node sampling after the control paths have been split.
\<close>

text \<open>
  Typing nonces rather than nodes also addresses the deficiency of key safety in CoSP that the
  same nonce may be contructed in different nodes (with different IDs) and therefore used in
  different ways. For example, we can have two nonce computations with IDs @{term id1}
  and @{term id2} on the same path and @{term id1} is used as randomness and @{term id2} by some
  output node. Thus, we are handing the randomness used in encryption over to the
  adversary. So, the simulator @{text sim2} below cannot push the randomness into the oracle
  and never access it, because the adversary would notice.
\<close>



datatype secret_key_type = SIGKEY | DKEY

datatype node_type =
    NonceT nonce_type
  | Secret_Key secret_key_type
  | PublicT

type_synonym protocol_path = "node_type list"


context fixes \<Gamma> :: "nonce_typing" begin
fun node_type_of_c :: "(constructor, nonce, destructor) computation \<Rightarrow> node_type"
where
  "node_type_of_c (Constructor EK_C) = PublicT"
| "node_type_of_c (Constructor DK_C) = Secret_Key DKEY"
| "node_type_of_c (Constructor VK_C) = PublicT"
| "node_type_of_c (Constructor SK_C) = Secret_Key SIGKEY"
| "node_type_of_c (Constructor E_C) = PublicT"
| "node_type_of_c (Constructor SIG_C) = PublicT"
| "node_type_of_c (Constructor PAIR_C) = PublicT"
| "node_type_of_c (Constructor EMPTY_C) = PublicT"
| "node_type_of_c (Constructor (STRING_C _)) = PublicT"
| "node_type_of_c (Constructor garbage_C) = PublicT"
| "node_type_of_c (Constructor garbageE_C) = PublicT"
| "node_type_of_c (Constructor garbageSIG_C) = PublicT"
| "node_type_of_c (Nonce (PNonce n)) = (case \<Gamma> n of Some nt \<Rightarrow> NonceT nt | None \<Rightarrow> PublicT)"
| "node_type_of_c (Nonce (ANonce _)) = PublicT"
| "node_type_of_c (Destructor _) = PublicT"
end

context fixes \<Gamma> :: nonce_typing begin
fun node_type_of_t :: "t \<Rightarrow> node_type"
where
  "node_type_of_t (EK _) = PublicT"
| "node_type_of_t (DK _) = Secret_Key DKEY"
| "node_type_of_t (VK _) = PublicT"
| "node_type_of_t (SK _) = Secret_Key SIGKEY"
| "node_type_of_t (E _ _ _) = PublicT"
| "node_type_of_t (SIG _ _ _) = PublicT"
| "node_type_of_t (PAIR _ _) = PublicT"
| "node_type_of_t EMPTY = PublicT"
| "node_type_of_t (STRING _ _) = PublicT"
| "node_type_of_t (garbage _) = PublicT"
| "node_type_of_t (garbageE _ _) = PublicT"
| "node_type_of_t (garbageSIG _ _) = PublicT"
| "node_type_of_t (N (PNonce n)) = (case \<Gamma> n of Some nt \<Rightarrow> NonceT nt | None \<Rightarrow> PublicT)"
| "node_type_of_t (N (ANonce _)) = PublicT"
end

lemma node_type_of_t_eq_NonceT [simp]:
  "node_type_of_t \<Gamma> t = NonceT nt \<longleftrightarrow> (\<exists>n. t = N (PNonce n) \<and> \<Gamma> n = Some nt)"
by(cases t rule: node_type_of_t.cases)(simp_all split: option.split)

inductive key_safe_computation :: "nonce_typing \<Rightarrow> protocol_path 
  \<Rightarrow> (constructor, nonce, destructor) computation \<Rightarrow> backlink list \<Rightarrow> bool"
  for \<Gamma> \<pi>
where
  EK: "\<pi> ! bl = NonceT Encryption \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Constructor EK_C) [bl]"
| DK: "\<pi> ! bl = NonceT Encryption \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Constructor DK_C) [bl]"
| VK: "\<pi> ! bl = NonceT Signature \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Constructor VK_C) [bl]"
| SK: "\<pi> ! bl = NonceT Signature \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Constructor SK_C) [bl]"
| E:
  "\<lbrakk> \<pi> ! bl1 = PublicT; \<pi> ! bl2 = PublicT; \<pi> ! bl3 = NonceT (Randomness (length \<pi>)) \<rbrakk>
  \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Constructor E_C) [bl1, bl2, bl3]"
| SIG: 
  "\<lbrakk> \<pi> ! bl1 = Secret_Key SIGKEY; \<pi> ! bl2 = PublicT; \<pi> ! bl3 = NonceT (Randomness (length \<pi>)) \<rbrakk>
  \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Constructor SIG_C) [bl1, bl2, bl3]"
| PAIR:
  "\<lbrakk> \<pi> ! bl1 = PublicT; \<pi> ! bl2 = PublicT \<rbrakk>
  \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Constructor PAIR_C) [bl1, bl2]"
| EMPTY:
  "key_safe_computation \<Gamma> \<pi> (Constructor EMPTY_C) []"
| STRING:
  "\<pi> ! bl = PublicT \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Constructor (STRING_C b)) [bl]"

| Nonce_secret:
  "\<Gamma> n = Some nt \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Nonce (PNonce n)) []"

| Nonce_public: -- \<open>Nonces may be used for other computations, too, but then they cannot be used as
    randomness for keys, encryptions, and signatures\<close>
  "\<Gamma> n = None \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Nonce (PNonce n)) []"

-- \<open>Adversary's nonces are not used (prot. cond. 9), so we cannot type them\<close>

| D: -- \<open>prot. cond. 6\<close>
  "\<lbrakk> \<pi> ! bl1 = Secret_Key DKEY; \<pi> ! bl2 = PublicT \<rbrakk>
  \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Destructor D) [bl1, bl2]"
| isenc: "\<pi> ! bl = PublicT \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Destructor isenc) [bl]"
| isek: "\<pi> ! bl = PublicT \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Destructor isek) [bl]"
| ekof: "\<pi> ! bl = PublicT \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Destructor ekof) [bl]"
| verify: "\<lbrakk> \<pi> ! bl1 = PublicT; \<pi> ! bl2 = PublicT \<rbrakk> \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Destructor verify) [bl1, bl2]"
| issig: "\<pi> ! bl = PublicT \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Destructor issig) [bl]"
| isvk: "\<pi> ! bl = PublicT \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Destructor isvk) [bl]"
| vkof: "\<pi> ! bl = PublicT \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Destructor vkof) [bl]"
| FST: "\<pi> ! bl = PublicT \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Destructor FST) [bl]"
| SND: "\<pi> ! bl = PublicT \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Destructor SND) [bl]"
| unstring: "\<pi> ! bl = PublicT \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Destructor (unstring b)) [bl]"
| equals: "\<lbrakk> \<pi> ! bl1 = PublicT; \<pi> ! bl2 = PublicT \<rbrakk> \<Longrightarrow> key_safe_computation \<Gamma> \<pi> (Destructor equals) [bl1, bl2]"

inductive_simps key_safe_computation_simps:
  "key_safe_computation \<Gamma> \<pi> (Constructor EK_C) args"
  "key_safe_computation \<Gamma> \<pi> (Constructor DK_C) args"
  "key_safe_computation \<Gamma> \<pi> (Constructor VK_C) args"
  "key_safe_computation \<Gamma> \<pi> (Constructor SK_C) args"
  "key_safe_computation \<Gamma> \<pi> (Constructor E_C) args"
  "key_safe_computation \<Gamma> \<pi> (Constructor SIG_C) args"
  "key_safe_computation \<Gamma> \<pi> (Constructor PAIR_C) args"
  "key_safe_computation \<Gamma> \<pi> (Constructor EMPTY_C) args"
  "key_safe_computation \<Gamma> \<pi> (Constructor (STRING_C b)) args"
  "key_safe_computation \<Gamma> \<pi> (Constructor garbage_C) args"
  "key_safe_computation \<Gamma> \<pi> (Constructor garbageE_C) args"
  "key_safe_computation \<Gamma> \<pi> (Constructor garbageSIG_C) args"
  "key_safe_computation \<Gamma> \<pi> (Nonce (PNonce n)) args"
  "key_safe_computation \<Gamma> \<pi> (Nonce (ANonce n') ) args"
  "key_safe_computation \<Gamma> \<pi> (Destructor D) args"
  "key_safe_computation \<Gamma> \<pi> (Destructor isenc) args"
  "key_safe_computation \<Gamma> \<pi> (Destructor isek) args"
  "key_safe_computation \<Gamma> \<pi> (Destructor ekof) args"
  "key_safe_computation \<Gamma> \<pi> (Destructor verify) args"
  "key_safe_computation \<Gamma> \<pi> (Destructor issig) args"
  "key_safe_computation \<Gamma> \<pi> (Destructor isvk) args"
  "key_safe_computation \<Gamma> \<pi> (Destructor vkof) args"
  "key_safe_computation \<Gamma> \<pi> (Destructor FST) args"
  "key_safe_computation \<Gamma> \<pi> (Destructor SND) args"
  "key_safe_computation \<Gamma> \<pi> (Destructor (unstring b)) args"
  "key_safe_computation \<Gamma> \<pi> (Destructor equals) args"

coinductive key_safe :: "nonce_typing \<Rightarrow> protocol_path \<Rightarrow> 'id cosp_protocol \<Rightarrow> bool"
  ("_, _/ \<turnstile> (_) \<surd>" [120, 0, 0] 65)          
  for \<Gamma>
where
  Computation:
  "\<lbrakk> key_safe_computation \<Gamma> \<pi> comput data; \<Gamma>, \<pi> @ [node_type_of_c \<Gamma> comput] \<turnstile> yes \<surd>; \<Gamma>, \<pi> \<turnstile> no \<surd>;
     \<And>bl. bl \<in> set data \<Longrightarrow> bl < length \<pi> \<rbrakk>               
  \<Longrightarrow> \<Gamma>, \<pi> \<turnstile> Computation ident comput data yes no \<surd>"
| Output:
  "\<lbrakk> \<Gamma>, \<pi> \<turnstile> succ \<surd>; \<pi> ! data = PublicT; data < length \<pi> \<rbrakk> \<Longrightarrow> \<Gamma>, \<pi> \<turnstile> Output ident data succ \<surd>"
| Input:
  "\<Gamma>, \<pi> @ [PublicT] \<turnstile> succ \<surd> \<Longrightarrow> \<Gamma>, \<pi> \<turnstile> Input ident succ \<surd>"
| Control:
  "(\<And>x. x \<in> neset A \<Longrightarrow> \<Gamma>, \<pi> \<turnstile> succs x \<surd>) \<Longrightarrow> \<Gamma>, \<pi> \<turnstile> Control ident out A succs \<surd>"
| Probabilistic:
  "(\<And>x. x \<in> set_pmf p \<Longrightarrow> \<Gamma>, \<pi> \<turnstile> succs x \<surd> ) \<Longrightarrow> \<Gamma>, \<pi> \<turnstile> Probabilistic ident p succs \<surd>"

inductive_cases key_safe_cases:
  "\<Gamma>, \<pi> \<turnstile> Computation ident comput data yes no \<surd>"
  "\<Gamma>, \<pi> \<turnstile> Output ident data succ \<surd>"
  "\<Gamma>, \<pi> \<turnstile> Input ident succ \<surd>"
  "\<Gamma>, \<pi> \<turnstile> Control ident out A succs \<surd>"
  "\<Gamma>, \<pi> \<turnstile> Probabilistic ident p succs \<surd>"

inductive_simps key_safe_simps:
  "\<Gamma>, \<pi> \<turnstile> Computation ident comput data yes no \<surd>"
  "\<Gamma>, \<pi> \<turnstile> Output ident data succ \<surd>"
  "\<Gamma>, \<pi> \<turnstile> Input ident succ \<surd>"
  "\<Gamma>, \<pi> \<turnstile> Control ident out A succs \<surd>"
  "\<Gamma>, \<pi> \<turnstile> Probabilistic ident p succs \<surd>"

(* text \<open>
  Randomness nonces must be unique. That is, they may only be used for encrypting single term.
  @{text rand2term} maps randomness a nonce to the key nonce and message whose encryption uses the nonce.
\<close>
type_synonym rand2term = "protocol_nonce \<Rightarrow> protocol_nonce \<times> t" *)

text \<open>
  Adversary nonces are used to store the serialisations of terms generated by the adversary.
  Therefore, we must require that the bitstrings have the correct shape.
\<close>

text \<open>@{text key_safe_t} characterises terms that may arise as terms during the execution of key-safe protocols\<close>
inductive key_safe_t
  :: "'more tagging_scheme \<Rightarrow> security \<Rightarrow> nonce_typing \<Rightarrow> t \<Rightarrow> node_type \<Rightarrow> bool" 
  ("_,_,_/ \<turnstile>t (_) : (_)" [120, 0, 0, 0, 120] 65)
  for A \<eta> \<Gamma>
where
  EK: "A,\<eta>,\<Gamma> \<turnstile>t t : NonceT Encryption \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t EK t : PublicT"
| EK_ANonce : "\<T>_ekey A \<eta> n \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t EK (N (ANonce n)) : PublicT"
| DK: "A,\<eta>,\<Gamma> \<turnstile>t t : NonceT Encryption \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t DK t : Secret_Key DKEY"
| VK: "A,\<eta>,\<Gamma> \<turnstile>t t : NonceT Signature \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t VK t : PublicT"
| VK_ANonce: "\<T>_vkey A \<eta> m \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t VK (N (ANonce m)): PublicT"
| SK: "A,\<eta>,\<Gamma> \<turnstile>t t : NonceT Signature \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t SK t : Secret_Key SIGKEY"
| E:
  "\<lbrakk> A,\<eta>,\<Gamma> \<turnstile>t t1 : PublicT; A,\<eta>,\<Gamma> \<turnstile>t t2 : PublicT; A,\<eta>,\<Gamma> \<turnstile>t t3 : NonceT (Randomness bl) \<rbrakk>
  \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t E t1 t2 t3 : PublicT"
| E_ANonce:
  "\<lbrakk> \<Gamma> n = Some Encryption; A,\<eta>,\<Gamma> \<turnstile>t t2 : PublicT; \<T>_cipher A \<eta> m \<rbrakk> 
  \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t E (EK (N (PNonce n))) t2 (N (ANonce m)) : PublicT" 
| SIG:
  "\<lbrakk> A,\<eta>,\<Gamma> \<turnstile>t t1 : Secret_Key SIGKEY; A,\<eta>,\<Gamma> \<turnstile>t t2 : PublicT; A,\<eta>,\<Gamma> \<turnstile>t t3 : NonceT (Randomness bl) \<rbrakk>
  \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t SIG t1 t2 t3 : PublicT" 
| SIG_ANonce: 
  "\<lbrakk> case n of PNonce n' \<Rightarrow> \<Gamma> n' = Some Signature | ANonce m \<Rightarrow> \<T>_vkey A \<eta> m;
     A,\<eta>,\<Gamma> \<turnstile>t t2 : PublicT; \<T>_signature A \<eta> m \<rbrakk>
  \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t SIG (SK (N n)) t2 (N (ANonce m)) : PublicT"
| PAIR:
  "\<lbrakk> A,\<eta>,\<Gamma> \<turnstile>t t1 : PublicT; A,\<eta>,\<Gamma> \<turnstile>t t2 : PublicT \<rbrakk>
  \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t PAIR t1 t2 : PublicT"
| EMPTY:
  "A,\<eta>,\<Gamma> \<turnstile>t EMPTY : PublicT"
| STRING:
  "A,\<eta>,\<Gamma> \<turnstile>t t : PublicT \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t STRING b t : PublicT"
| Nonce_secret:
  "\<Gamma> n = Some nt \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t N (PNonce n) : NonceT nt"
| Nonce_public:
  "\<Gamma> n = None \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t N (PNonce n) : PublicT"
| ANonce: "\<T>_nonce A \<eta> m \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t N (ANonce m) : PublicT"
| garbage: "A,\<eta>,\<Gamma> \<turnstile>t garbage t : PublicT"
| garbageE:
  "\<lbrakk> A,\<eta>,\<Gamma> \<turnstile>t t : PublicT; \<T>_cipher A \<eta> m \<rbrakk> \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t garbageE t (N (ANonce m)) : PublicT"  
| garbageSIG:
  "\<lbrakk> A,\<eta>,\<Gamma> \<turnstile>t t : PublicT; \<T>_signature A \<eta> m \<rbrakk> \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t garbageSIG t (N (ANonce m)) : PublicT"

lemma key_safe_t_nodetype: "A,\<eta>,\<Gamma> \<turnstile>t t : nt \<Longrightarrow> node_type_of_t \<Gamma> t = nt"
by(induction rule: key_safe_t.induct)(auto)

inductive_simps key_safe_t_simps [simp]:
  "A,\<eta>,\<Gamma> \<turnstile>t E t1 t2 t3 : nt"
  "A,\<eta>,\<Gamma> \<turnstile>t EK t : nt"
  "A,\<eta>,\<Gamma> \<turnstile>t DK t : nt"
  "A,\<eta>,\<Gamma> \<turnstile>t SIG t1 t2 t3 : nt"
  "A,\<eta>,\<Gamma> \<turnstile>t VK t : nt"
  "A,\<eta>,\<Gamma> \<turnstile>t SK t : nt"
  "A,\<eta>,\<Gamma> \<turnstile>t PAIR t1 t2 : nt"
  "A,\<eta>,\<Gamma> \<turnstile>t EMPTY : nt"
  "A,\<eta>,\<Gamma> \<turnstile>t STRING b t : nt"
  "A,\<eta>,\<Gamma> \<turnstile>t N nonce : nt"
  "A,\<eta>,\<Gamma> \<turnstile>t garbage t : nt"
  "A,\<eta>,\<Gamma> \<turnstile>t garbageE t1 t2 : nt"
  "A,\<eta>,\<Gamma> \<turnstile>t garbageSIG t1 t2 : nt"

inductive_simps key_safe_t_NonceT_iff: "A,\<eta>,\<Gamma> \<turnstile>t t : NonceT nt"

method key_safe_hybrid_step_cases uses r =
  (auto, (simp_all split: if_split_asm option.split_asm)?, (rule r, (simp+)?, (auto simp add: in_set_conv_nth)?)+)

lemma key_safe_hybrid_step1_cases:
  assumes step: "Some (Some t, (s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step1' sim ((s,v,f),\<sigma>))"
  and ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v"
  notes [simp] = key_safe_simps key_safe_computation_simps
  obtains (Nonce) id n no 
    where "v = Computation id (Nonce (PNonce n)) [] v' no" "t = N (PNonce n)" "s' = s"
      "f' = f @ [N (PNonce n)]" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [node_type_of_t \<Gamma> (N (PNonce n))] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (E_C) id no bl1 bl2 bl3 n k
    where "v = Computation id (Constructor E_C) [bl1, bl2, bl3] v' no"
      "bl1 < length f" "bl2 < length f" "bl3 < length f"
      "EK (N k) \<in> set f" "f ! bl2 \<in> set f" "N (PNonce n) \<in> set f"
      "node_type_of_t \<Gamma> (f ! bl2) = PublicT" "f ! bl3 = N (PNonce n)" "\<Gamma> n = Some (Randomness (length f))"
      "\<T> (f ! bl2)" "f ! bl1 = EK (N k)" "t = E (EK (N k)) (f ! bl2) (N (PNonce n))"
      "s' = s" "f' = f @ [E (EK (N k)) (f ! bl2) (N (PNonce n))]" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (EK_C) id no bl n
    where "v = Computation id (Constructor EK_C) [bl] v' no"
      "bl < length f" "f ! bl = N (PNonce n)" "N (PNonce n) \<in> set f"
      "\<Gamma> n = Some Encryption" "t = EK (N (PNonce n))" "s' = s"
      "f' = f @ [EK (N (PNonce n))]" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (DK_C) id no bl n
    where "v = Computation id (Constructor DK_C) [bl] v' no"
      "bl < length f" "f ! bl = N (PNonce n)" "N (PNonce n) \<in> set f"
      "\<Gamma> n = Some Encryption" "t = DK (N (PNonce n))" "s' = s" "f' = f @ [DK (N (PNonce n))]" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [Secret_Key DKEY] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (SIG_C) id no bl1 bl2 bl3 n k
    where "v = Computation id (Constructor SIG_C) [bl1, bl2, bl3] v' no"
      "bl1 < length f" "bl2 < length f" "bl3 < length f"
      "f ! bl1 = SK (N k)" "f ! bl3 = N (PNonce n)" "node_type_of_t \<Gamma> (f ! bl2) = PublicT"
      "SK (N k) \<in> set f" "N (PNonce n) \<in> set f" "f ! bl2 \<in> set f"
      "\<Gamma> n = Some (Randomness (length f))" "\<T> (f ! bl2)"
      "t = SIG (SK (N k)) (f ! bl2) (N (PNonce n))"
      "s' = s" "f' = f @ [SIG (SK (N k)) (f ! bl2) (N (PNonce n))]" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (VK_C) id no bl n
    where "v = Computation id (Constructor VK_C) [bl] v' no"
      "bl < length f" "f ! bl = N (PNonce n)" "N (PNonce n) \<in> set f"
      "\<Gamma> n = Some Signature" "t = VK (N (PNonce n))" "s' = s" "f' = f @ [VK (N (PNonce n))]" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (SK_C) id no bl n
     where "v = Computation id (Constructor SK_C) [bl] v' no"
       "bl < length f" "f ! bl = N (PNonce n)" "N (PNonce n) \<in> set f"
       "\<Gamma> n = Some Signature" "t = SK (N (PNonce n))" "s' = s" "f' = f @ [SK (N (PNonce n))]" "\<sigma>' = \<sigma>"
       "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [Secret_Key SIGKEY] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (PAIR_C) id no bl1 bl2
    where "v = Computation id (Constructor PAIR_C) [bl1, bl2] v' no"
      "bl1 < length f" "bl2 < length f" "f ! bl1 \<in> set f" "f ! bl2 \<in> set f"
      "node_type_of_t \<Gamma> (f ! bl1) = PublicT" "node_type_of_t \<Gamma> (f ! bl2) = PublicT"
      "\<T> (f ! bl1)" "\<T> (f ! bl2)" "t = PAIR (f ! bl1) (f ! bl2)"
      "s' = s" "f' = f @ [PAIR (f ! bl1) (f ! bl2)]" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (EMPTY_C) id no
    where "v = Computation id (Constructor EMPTY_C) [] v' no"
      "t = EMPTY" "s' = s" "f' = f @ [EMPTY]" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (STRING_C) id no b bl
    where "v = Computation id (Constructor (STRING_C b)) [bl] v' no"
      "bl < length f" "f ! bl \<in> set f" "\<S> (f ! bl)" "t = STRING b (f ! bl)"
      "node_type_of_t \<Gamma> (f ! bl) = PublicT" "s' = s" "f' = f @ [STRING b (f ! bl)]" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (D) id no k r bl1 bl2
    where "v = Computation id (Destructor D) [bl1, bl2] v' no"
      "s' = s" "f ! bl1 = DK k" "bl1 < length f" "f ! bl2 = E (EK k) t r" "bl2 < length f"
      "DK k \<in> set f" "E (EK k) t r \<in> set f" "f' = f @ [t]" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (isenc) id no bl t1 t2 t3
    where "v = Computation id (Destructor isenc) [bl] v' no" "s' = s"
      "t = E (EK t1) t2 t3 \<or> t = garbageE t1 t2" "f' = f @ [t]"
      "bl < length f" "f ! bl = t" "t \<in> set f" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (isek) id no t' bl
    where "v = Computation id (Destructor isek) [bl] v' no"
      "t = EK t'" "s' = s" "f ! bl = EK t'" "bl < length f" "EK t' \<in> set f"
      "f' = f @ [EK t']" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (ekof) id no t1 t2 t3 bl
    where "v = Computation id (Destructor ekof) [bl] v' no"
      "s' = s" "bl < length f" "f' = f @ [t]" "\<sigma>' = \<sigma>"
      "t = EK t1 \<and> f ! bl = E (EK t1) t2 t3 \<and> E (EK t1) t2 t3 \<in> set f \<or>
       t = t1 \<and> f ! bl = garbageE t1 t2 \<and> garbageE t1 t2 \<in> set f"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (verify) id no k r bl1 bl2
    where "v = Computation id (Destructor verify) [bl1, bl2] v' no" "s' = s"
      "f ! bl1 = VK k" "bl1 < length f" "VK k \<in> set f"
      "f ! bl2 = SIG (SK k) t r" "bl2 < length f" "SIG (SK k) t r \<in> set f"
      "f' = f @ [t]" "\<sigma>' = \<sigma>" 
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (issig) id no bl t1 t2 t3
    where "v = Computation id (Destructor issig) [bl] v' no" "s' = s"
      "bl < length f" "f' = f @ [t]" "\<sigma>' = \<sigma>" "f ! bl = t" "t \<in> set f"
      "t = SIG (SK t1) t2 t3 \<or> t = garbageSIG t1 t2"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (isvk) id no bl t'
     where "v = Computation id (Destructor isvk) [bl] v' no"
       "t = VK t'" "s' = s" "f ! bl = VK t'" "bl < length f" "VK t' \<in> set f" "f' = f @ [VK t']" "\<sigma>' = \<sigma>"
       "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (vkof) id no bl t1 t2 t3
    where "v = Computation id (Destructor vkof) [bl] v' no" "s' = s" "f' = f @ [t]" "\<sigma>' = \<sigma>"
      "f ! bl = SIG (SK t1) t2 t3 \<and> SIG (SK t1) t2 t3 \<in> set f \<and> t = VK t1 \<or>
       f ! bl = garbageSIG t1 t2 \<and> garbageSIG t1 t2 \<in> set f \<and> t = t1"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (FST) id no bl t2
    where "v = Computation id (Destructor FST) [bl] v' no" "s' = s"
      "bl < length f" "f ! bl = PAIR t t2" "PAIR t t2 \<in> set f" "f' = f @ [t]" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (SND) id no bl t1
    where "v = Computation id (Destructor SND) [bl] v' no" "s' = s"
      "bl < length f" "f ! bl = PAIR t1 t" "PAIR t1 t \<in> set f" "f' = f @ [t]" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (unstring) id no bl b
    where "v = Computation id (Destructor (unstring b)) [bl] v' no" "s' = s"
      "bl < length f" "f ! bl = STRING b t""STRING b t \<in> set f" "f' = f @ [t]" "\<sigma>' = \<sigma>"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (equals) id no bl1 bl2
    where "v = Computation id (Destructor equals) [bl1, bl2] v' no" "s' = s"
      "bl1 < length f" "f ! bl1 \<in> set f" "bl2 < length f" "t = f ! bl1"
      "node_type_of_t \<Gamma> (f ! bl1) = PublicT" "f' = f @ [f ! bl1]" "\<sigma>' = \<sigma>" "f ! bl2 = f ! bl1"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" "\<Gamma>, map (node_type_of_t \<Gamma>) f \<turnstile> no \<surd>"
  | (Input) id
    where "v = Input id v'" "s' = s" "Some (t, \<sigma>') \<in> results'_gpv (gen_message sim \<sigma>)" "f' = f @ [t]"
      "\<Gamma>, map (node_type_of_t \<Gamma>) f @ [PublicT] \<turnstile> v' \<surd>" 
proof(cases v)
  case [simp]: (Computation id comp c_data yes no)
  show ?thesis
  proof(cases comp)
    case *: (Nonce n)
    thus ?thesis using step ks by(cases n)(auto intro: Nonce)
  next
    case [simp]: (Constructor c)
    show ?thesis
    proof(cases c)
      { case *: E_C       thus ?thesis using ks step by-(key_safe_hybrid_step_cases r: E_C) }
      { case *: EK_C      thus ?thesis using ks step by-(key_safe_hybrid_step_cases r: EK_C) }
      { case *: DK_C      thus ?thesis using ks step by-(key_safe_hybrid_step_cases r: DK_C) }
      { case *: SIG_C     thus ?thesis using ks step by-(key_safe_hybrid_step_cases r: SIG_C) }
      { case *: VK_C      thus ?thesis using ks step by-(key_safe_hybrid_step_cases r: VK_C) }
      { case *: SK_C      thus ?thesis using ks step by-(key_safe_hybrid_step_cases r: SK_C) }
      { case *: PAIR_C    thus ?thesis using ks step by-(key_safe_hybrid_step_cases r: PAIR_C) }
      { case *: EMPTY_C   thus ?thesis using ks step by-(key_safe_hybrid_step_cases r: EMPTY_C) }
      { case *: STRING_C  thus ?thesis using ks step by-(key_safe_hybrid_step_cases r: STRING_C) }
      { case garbage_C    thus ?thesis using ks by auto }
      { case garbageE_C   thus ?thesis using ks by auto }
      { case garbageSIG_C thus ?thesis using ks by auto }
    qed
  next
    case [simp]: (Destructor d)
    hence t: "\<D> d (map (op ! f) c_data) = Some t" using step Computation by(simp split: option.split_asm)
    note [elim!] = \<D>s_SomeE
    show ?thesis
    proof(cases d)
      { case *: D        thus ?thesis using ks t step by-(key_safe_hybrid_step_cases r: D) }
      { case *: isenc    thus ?thesis using ks t step by-(key_safe_hybrid_step_cases r: isenc) }
      { case *: isek     thus ?thesis using ks t step by-(key_safe_hybrid_step_cases r: isek) } 
      { case *: ekof     thus ?thesis using ks t step by-(key_safe_hybrid_step_cases r: ekof) }  
      { case *: verify   thus ?thesis using ks t step by-(key_safe_hybrid_step_cases r: verify) }   
      { case *: issig    thus ?thesis using ks t step by-(key_safe_hybrid_step_cases r: issig) }
      { case *: isvk     thus ?thesis using ks t step by-(key_safe_hybrid_step_cases r: isvk) }
      { case *: vkof     thus ?thesis using ks t step by-(key_safe_hybrid_step_cases r: vkof) } 
      { case *: FST      thus ?thesis using ks t step by-(key_safe_hybrid_step_cases r: FST) }
      { case *: SND      thus ?thesis using ks t step by-(key_safe_hybrid_step_cases r: SND) } 
      { case *: unstring thus ?thesis using ks t step by-(key_safe_hybrid_step_cases r: unstring) } 
      { case *: equals   thus ?thesis using ks t step by-(key_safe_hybrid_step_cases r: equals) } 
    qed
  qed
next
  case Output with step show ?thesis by(auto simp add: gpv.set_map)
next
  case *: Input with step ks show ?thesis by(auto split: if_split_asm intro: Input simp add: gpv.set_map Some_eq_bind_conv)
next
  case Control with step show ?thesis by(auto simp add: gpv.set_map)
next
  case Probabilistic with step show ?thesis by(auto)
qed

definition key_safe_trace :: "'more tagging_scheme \<Rightarrow> security \<Rightarrow> nonce_typing \<Rightarrow> t list \<Rightarrow> bool"
where "key_safe_trace A \<eta> \<Gamma> f \<longleftrightarrow> (\<forall>t \<in> set f. A,\<eta>,\<Gamma> \<turnstile>t t : node_type_of_t \<Gamma> t)"

lemma key_safe_traceD: "\<lbrakk> key_safe_trace A \<eta> \<Gamma> f; i < length f \<rbrakk> \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t f ! i : node_type_of_t \<Gamma> (f ! i)"
by(simp add: key_safe_trace_def)

lemma key_safe_traceD': "\<lbrakk> key_safe_trace A \<eta> \<Gamma> f; t \<in> set f \<rbrakk> \<Longrightarrow>  A,\<eta>,\<Gamma> \<turnstile>t t : node_type_of_t \<Gamma> t"
by(simp add: key_safe_trace_def)

lemma key_safe_trace_Nil [simp]: "key_safe_trace A \<eta> \<Gamma> []"
by(simp add: key_safe_trace_def)

lemma key_safe_trace_Cons [simp]:
  "key_safe_trace A \<eta> \<Gamma> (t # f) \<longleftrightarrow> A,\<eta>,\<Gamma> \<turnstile>t t : node_type_of_t \<Gamma> t \<and> key_safe_trace A \<eta> \<Gamma> f"
by(simp add: key_safe_trace_def)

lemma key_safe_trace_append [simp]:
  "key_safe_trace A \<eta> \<Gamma> (f @ f') \<longleftrightarrow> key_safe_trace A \<eta> \<Gamma> f \<and> key_safe_trace A \<eta> \<Gamma> f'"
by(auto simp add: key_safe_trace_def)

lemma key_safe_t_hybrid_step1:
  assumes step: "Some (Some t, (s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step1' sim ((s,v,f),\<sigma>))"
  and ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v"
  and kst: "key_safe_trace A \<eta> \<Gamma> f"
  and gen_message: "Some (t, \<sigma>') \<in> results'_gpv (gen_message sim \<sigma>) \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t t : node_type_of_t \<Gamma> t"
  shows "A,\<eta>,\<Gamma> \<turnstile>t t : node_type_of_t \<Gamma> t"
using step ks
proof(cases rule: key_safe_hybrid_step1_cases)
  note [simp del] = nth_mem and [dest!] = key_safe_traceD'[OF kst]
  { case Nonce    thus ?thesis by(auto split: option.split) }
  { case E_C      thus ?thesis by(auto) }
  { case EK_C     thus ?thesis by(auto) }
  { case DK_C     thus ?thesis by(auto) }
  { case SIG_C    thus ?thesis by(auto) }
  { case VK_C     thus ?thesis by(auto) }
  { case SK_C     thus ?thesis by(auto) }
  { case PAIR_C   thus ?thesis by(auto) }
  { case EMPTY_C  thus ?thesis by(auto) }
  { case STRING_C thus ?thesis by(auto) }
  { case D        thus ?thesis by(auto simp add: key_safe_t_nodetype) }
  { case isenc    thus ?thesis by(auto) }
  { case isek     thus ?thesis by(auto) }
  { case ekof     thus ?thesis by(auto simp add: key_safe_t_nodetype) }
  { case verify   thus ?thesis by(auto simp add: key_safe_t_nodetype) }
  { case issig    thus ?thesis by(auto) }
  { case isvk     thus ?thesis by(auto) }
  { case vkof     thus ?thesis by(auto simp add: key_safe_t_nodetype)(simp split: nonce.split_asm) }
  { case FST      thus ?thesis by(auto simp add: key_safe_t_nodetype) }
  { case SND      thus ?thesis by(auto simp add: key_safe_t_nodetype) }
  { case unstring thus ?thesis by(auto simp add: key_safe_t_nodetype) }
  { case equals   thus ?thesis by(clarsimp split: if_split_asm) }
  { case Input    thus ?thesis by(auto dest: gen_message) }
qed

lemma key_safe_trace_hybrid_step:
  assumes step: "Some ((s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step' sim ((s,v,f),\<sigma>))"
  and ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v"
  and kst: "key_safe_trace A \<eta> \<Gamma> f"
  and gen_message: "\<And>t \<sigma>'. Some (t, \<sigma>') \<in> results'_gpv (gen_message sim \<sigma>) \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t t : node_type_of_t \<Gamma> t"
  shows "key_safe_trace A \<eta> \<Gamma> f'"
proof -
  from step obtain to \<sigma>' where "Some (to, (s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step1' sim ((s, v, f), \<sigma>))"
    by(clarsimp simp add: hybrid_step_def gpv.set_map split: option.split_asm)
  with key_safe_t_hybrid_step1[OF _ ks kst gen_message, of _ s' v' f' \<sigma>' sim s \<sigma>] kst
  show ?thesis
    apply(cases v)
    apply(auto simp add: split_def gpv.set_map Some_eq_bind_conv split: if_split_asm computation.split_asm option.split_asm)
    apply(auto elim!: meta_allE meta_impE intro: rev_image_eqI)
    done
qed

lemma key_safe_hybrid_step:
  assumes step: "Some ((s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step' sim ((s,v,f),\<sigma>))"
  and ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v"
  and kst: "key_safe_trace A \<eta> \<Gamma> f"
  and gen_message: "\<And>t \<sigma>'. Some (t, \<sigma>') \<in> results'_gpv (gen_message sim \<sigma>) \<Longrightarrow> node_type_of_t \<Gamma> t = PublicT"
  shows "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f') v'"
proof -
  from step obtain to \<sigma>''
    where step1: "Some (to, (s', v', f'), \<sigma>'') \<in> results'_gpv (hybrid_step1' sim ((s, v, f), \<sigma>))"
    by(auto simp add: hybrid_step_def gpv.set_map split: option.split_asm)
  show ?thesis
  proof(cases to)
    case None
    with ks step1 show ?thesis by(cases)(auto simp add: gpv.set_map Some_eq_bind_conv split: option.split_asm if_split_asm)
  next
    case (Some t)
    from step1 ks show ?thesis unfolding Some
      by(cases rule: key_safe_hybrid_step1_cases)(auto dest!: key_safe_traceD'[OF kst] gen_message simp add: key_safe_t_nodetype)
  qed
qed

subsection \<open>Randomness usage\<close>

text \<open>Randomness nonces may only be used to encrypt a single term. We define functions to associate 
  a nonce with all the terms in which it is used as randomness. Below, we will require that each nonce
  is associated with at most one such term.\<close>

datatype random_usage' = RUEncryption t t | RUSignature t t

type_synonym random_usage = "protocol_nonce \<Rightarrow> random_usage' set"
translations
  (type) "random_usage" \<leftharpoondown> (type) "nat \<Rightarrow> random_usage' set"
typ random_usage

abbreviation (input) empty_usage :: "random_usage"
where "empty_usage \<equiv> \<bottom>"

definition single_usage :: "protocol_nonce \<Rightarrow> random_usage' \<Rightarrow> random_usage"
where "single_usage r tt r' = (if r = r' then {tt} else {})"

lemma in_single_usage_iff [simp]: "tt' \<in> single_usage r tt r' \<longleftrightarrow> r = r' \<and> tt' = tt"
by(simp add: single_usage_def)

lemma single_usage_empty_iff [simp]: "single_usage r tt r' = {} \<longleftrightarrow> r \<noteq> r'"
by(simp add: single_usage_def)

lemma single_usage_leD:
  "single_usage r sru \<le> ru \<Longrightarrow> sru \<in> ru r"
by(simp add: le_fun_def single_usage_def split: if_split_asm)

fun random_usage :: "t \<Rightarrow> random_usage"
where
  "random_usage (N n) = empty_usage"
| "random_usage (E t1 t2 t3) = 
   random_usage t1 \<squnion> random_usage t2 \<squnion> 
   (case t3 of N (PNonce r) \<Rightarrow> single_usage r (RUEncryption t1 t2) | _ \<Rightarrow> empty_usage)" 
| "random_usage (EK t) = empty_usage" (* why? *)
| "random_usage (DK t) = empty_usage" (* why? *)
| "random_usage (SIG t1 t2 t3) =
   random_usage t1 \<squnion> random_usage t2 \<squnion>
   (case t3 of N (PNonce r) \<Rightarrow> single_usage r (RUSignature t1 t2) | _ \<Rightarrow> empty_usage)"
| "random_usage (VK t) = empty_usage" (* why? *)
| "random_usage (SK t) = empty_usage" (* why? *)
| "random_usage (PAIR t1 t2) = random_usage t1 \<squnion> random_usage t2"
| "random_usage EMPTY = empty_usage"
| "random_usage (STRING b t) = random_usage t"
| "random_usage (garbage t) = random_usage t"
| "random_usage (garbageE t1 t2) = random_usage t1 \<squnion> random_usage t2"
| "random_usage (garbageSIG t1 t2) = random_usage t1 \<squnion> random_usage t2"

context begin
interpretation sup_list: comm_monoid_list "op \<squnion> :: 'a :: bounded_semilattice_sup_bot \<Rightarrow> _" \<bottom>
  by unfold_locales

definition random_usages :: "t list \<Rightarrow> random_usage"
where "random_usages ts = sup_list.F (map random_usage ts)"

lemma random_usages_simps [simp, code]:
  "random_usages [] = empty_usage"
  "random_usages (t # ts) = random_usage t \<squnion> random_usages ts"
by(simp_all add: random_usages_def fun_eq_iff)

lemma random_usages_append [simp]:
  "random_usages (ts @ ts') = random_usages ts \<squnion> random_usages ts'"
by(simp add: random_usages_def)

lemma member_le_random_usages: "t \<in> set ts \<Longrightarrow> random_usage t \<le> random_usages ts"
by(induction ts)(auto simp add: le_supI2)

end

definition wf_random_usage' :: "nonce_typing \<Rightarrow> nat \<Rightarrow> random_usage \<Rightarrow> bool"
where "wf_random_usage' \<Gamma> n ru \<longleftrightarrow>
  (\<forall>r. \<forall>tt tt'. tt \<in> ru r \<longrightarrow> tt' \<in> ru r \<longrightarrow> tt = tt') \<and>
  (\<forall>r. ru r \<noteq> {} \<longrightarrow> (\<exists>bl. \<Gamma> r = Some (Randomness bl) \<and> bl < n))"

abbreviation wf_random_usage :: "nonce_typing \<Rightarrow> t list \<Rightarrow> bool"
where "wf_random_usage \<Gamma> ts \<equiv> wf_random_usage' \<Gamma> (length ts) (random_usages ts)"

lemma wf_random_usage'I [intro?]:
  "\<lbrakk> \<And>r tt tt'. \<lbrakk> tt \<in> ru r; tt' \<in> ru r \<rbrakk> \<Longrightarrow> tt = tt';
     \<And>r tt. tt \<in> ru r \<Longrightarrow> \<exists>bl. \<Gamma> r = Some (Randomness bl) \<and> bl < n \<rbrakk>
   \<Longrightarrow> wf_random_usage' \<Gamma> n ru"
by(auto simp add: wf_random_usage'_def)

lemma wf_random_usage'D:
  assumes "wf_random_usage' \<Gamma> n ru"
  shows wf_random_usage'D1: "tt \<in> ru r \<Longrightarrow> tt' \<in> ru r \<Longrightarrow> tt = tt'"
  and wf_random_usage'D2: "ru r \<noteq> {} \<Longrightarrow> \<exists>bl. \<Gamma> r = Some (Randomness bl) \<and> bl < n"
using assms by - ((cases tt; cases tt'), auto simp add: wf_random_usage'_def)

lemma wf_random_usage'_bot [simp]: "wf_random_usage' \<Gamma> n \<bottom>"
by(auto intro: wf_random_usage'I)

lemma wf_random_usage'_mono:
  "\<lbrakk> wf_random_usage' \<Gamma> n ru'; ru \<le> ru'; n \<le> n' \<rbrakk> \<Longrightarrow> wf_random_usage' \<Gamma> n' ru"
by(simp add: wf_random_usage'_def le_fun_def)(blast 11 intro: order.strict_trans2)

lemma wf_random_usage_Nil [simp]: "wf_random_usage \<Gamma> []"
by(simp add: wf_random_usage'_def)

lemma wf_random_usage'_single_usage_sup_iff [simp]:
  "wf_random_usage' \<Gamma> n (single_usage r tt \<squnion> ru) \<longleftrightarrow> 
   wf_random_usage' \<Gamma> n ru \<and> ru r \<subseteq> {tt} \<and> (\<exists>bl. \<Gamma> r = Some (Randomness bl) \<and> bl < n)"
by(auto 4 3 intro!: wf_random_usage'I dest: wf_random_usage'D1 elim: wf_random_usage'D2)

lemma wf_random_usage'_sup_single_usage_iff [simp]:
  "wf_random_usage' \<Gamma> n (ru \<squnion> single_usage r tt) \<longleftrightarrow> 
   wf_random_usage' \<Gamma> n ru \<and> ru r \<subseteq> {tt} \<and> (\<exists>bl. \<Gamma> r = Some (Randomness bl) \<and> bl < n)"
by(subst sup_commute)(rule wf_random_usage'_single_usage_sup_iff)

lemma wf_random_usage_hybrid_step1:
  assumes wfr: "wf_random_usage \<Gamma> f"
  and ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v"
  and gen: "\<And>t \<sigma>'. Some (t, \<sigma>') \<in> results'_gpv (gen_message sim \<sigma>) \<Longrightarrow> wf_random_usage \<Gamma> (f @ [t])"
  and step: "Some (to, (s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step1' sim ((s,v,f),\<sigma>))"
  shows "wf_random_usage \<Gamma> f'"
proof(cases v)
  case [simp]: (Computation id comp c_data yes no)
  show ?thesis
  proof(cases comp)
    case (Nonce n)
    thus ?thesis using step wfr by(auto intro: wf_random_usage'_mono)
  next
    case [simp]: (Destructor d)
    show ?thesis using step wfr
    proof(cases "\<D> d (map (op ! f) c_data)")
      case (Some t)
      with step have [simp]: "to = Some t" "s' = s" "v' = yes" "f' = f @ [t]" "\<sigma>' = \<sigma>" 
        and t: "\<D> d (map (op ! f) c_data) = Some t" by simp_all
      note [elim!] = \<D>s_SomeE
        and [simp] = key_safe_simps key_safe_computation_simps
      have "random_usage t \<le> random_usages f"
      proof(cases d)
        case [simp]: D
        from t ks obtain k r where "E (EK k) t r \<in> set f" by(fastforce simp add: in_set_conv_nth)
        from member_le_random_usages[OF this] show ?thesis by simp
      next
        case [simp]: isenc
        from t ks have "t \<in> set f" by(auto simp add: in_set_conv_nth)
        from member_le_random_usages[OF this] show ?thesis by simp
      next
        case [simp]: isek
        from t ks have "t \<in> set f" by(auto simp add: in_set_conv_nth)
        from member_le_random_usages[OF this] show ?thesis by simp
      next
        case [simp]: ekof
        from t ks consider m r where "E t m r \<in> set f"
          | t' where "garbageE t t' \<in> set f" by(fastforce simp add: in_set_conv_nth)
        thus ?thesis by cases(drule member_le_random_usages; simp; fail)+
      next
        case [simp]: verify
        from t ks obtain k r where "SIG (SK k) t r \<in> set f" by(fastforce simp add: in_set_conv_nth)
        from member_le_random_usages[OF this] show ?thesis by simp
      next
        case [simp]: issig
        from t ks have "t \<in> set f" by(auto simp add: in_set_conv_nth)
        from member_le_random_usages[OF this] show ?thesis by simp
      next
        case [simp]: isvk
        from t ks have "t \<in> set f" by(auto simp add: in_set_conv_nth)
        from member_le_random_usages[OF this] show ?thesis by simp
      next
        case [simp]: vkof
        from t ks consider k m r where "t = VK k" "SIG (SK k) m r \<in> set f"
          | t' where "garbageSIG t t' \<in> set f" by(fastforce simp add: in_set_conv_nth)
        thus ?thesis by cases(drule member_le_random_usages; simp; fail)+
      next
        case [simp]: FST
        from t ks obtain t' where "PAIR t t' \<in> set f" by(fastforce simp add: in_set_conv_nth)
        from member_le_random_usages[OF this] show ?thesis by simp
      next
        case [simp]: SND
        from t ks obtain t' where "PAIR t' t \<in> set f" by(fastforce simp add: in_set_conv_nth)
        from member_le_random_usages[OF this] show ?thesis by simp
      next
        case [simp]: (unstring b)
        from t ks have "STRING b t \<in> set f" by(fastforce simp add: in_set_conv_nth)
        from member_le_random_usages[OF this] show ?thesis by simp
      next
        case [simp]: equals
        from t ks have "t \<in> set f" by(fastforce simp add: in_set_conv_nth)
        from member_le_random_usages[OF this] show ?thesis by simp
      qed
      hence "random_usages (f @ [t]) = random_usages f" by(simp add: sup.absorb1)
      thus "wf_random_usage \<Gamma> f'" using wfr
        by(auto intro: wf_random_usage'_mono)
    qed simp
  next
    case [simp]: (Constructor C)
    note [simp] = key_safe_simps key_safe_computation_simps
      and [elim!] = \<C>s_E
    show ?thesis using step wfr
    proof(cases "\<T> (\<C> C (map (op ! f) c_data))")
      case \<T>: True
      with step have [simp]: "to = Some (\<C> C (map (op ! f) c_data))" (is "_ = Some ?t")
        "s' = s" "v' = yes" "f' = f @ [\<C> C (map (op ! f) c_data)]" "\<sigma>' = \<sigma>" by simp_all
      from \<T> show ?thesis
      proof(cases)
        case (E m k r)
        with ks obtain bl1 bl2 bl3 n where [simp]: "c_data = [bl1, bl2, bl3]" "C = E_C"
          and bl: "bl1 < length f" "bl2 < length f" "bl3 < length f"
          and r: "f ! bl3 = N (PNonce n)"
          and n: "\<Gamma> n = Some (Randomness (length f))" "r = PNonce n" by auto
        let ?RU = "single_usage n (RUEncryption (f ! bl1) (f ! bl2))"
        from wfr n have "random_usages f n = {}"
          by(auto simp add: dest: wf_random_usage'D2)
        then have "wf_random_usage' \<Gamma> (Suc (length f)) (?RU \<squnion> random_usages f)"
          using n wfr by(simp add: wf_random_usage'_mono)
        moreover from bl have "f ! bl1 \<in> set f" "f ! bl2 \<in> set f" by(auto simp add: in_set_conv_nth)
        hence "random_usage (f ! bl1) \<le> random_usages f" "random_usage (f ! bl2) \<le> random_usages f"
          by(blast intro: member_le_random_usages)+
        then have "random_usages (f @ [?t]) \<le> ?RU \<squnion> random_usages f"
          using r by (simp add: sup.coboundedI1 sup.commute)
        ultimately have "wf_random_usage \<Gamma> (f @ [?t])" by(rule wf_random_usage'_mono) simp
        then show ?thesis by simp
      next
        case (EK n)
        with wfr show ?thesis by(simp add: wf_random_usage'_mono)
      next
        case (DK n)
        with wfr show ?thesis by(simp add: wf_random_usage'_mono)
      next
        case (SIG k m r)
        with ks obtain bl1 bl2 bl3 n where [simp]: "c_data = [bl1, bl2, bl3]" "C = SIG_C"
          and bl: "bl1 < length f" "bl2 < length f" "bl3 < length f"
          and r: "f ! bl3 = N (PNonce n)"
          and n: "\<Gamma> n = Some (Randomness (length f))" "r = PNonce n" by auto
        let ?RU = "single_usage n (RUSignature (f ! bl1) (f ! bl2))"
        from wfr n have "random_usages f n = {}"
          by(auto simp add: dest: wf_random_usage'D2)
        then have "wf_random_usage' \<Gamma> (Suc (length f)) (?RU \<squnion> random_usages f)"
          using n wfr by(simp add: wf_random_usage'_mono)
        moreover from bl have "f ! bl1 \<in> set f" "f ! bl2 \<in> set f" by(auto simp add: in_set_conv_nth)
        hence "random_usage (f ! bl1) \<le> random_usages f" "random_usage (f ! bl2) \<le> random_usages f"
          by(blast intro: member_le_random_usages)+
        then have "random_usages (f @ [?t]) \<le> ?RU \<squnion> random_usages f"
          using r by (simp add: sup.coboundedI1 sup.commute)
        ultimately have "wf_random_usage \<Gamma> (f @ [?t])" by(rule wf_random_usage'_mono) simp
        then show ?thesis by simp
      next
        case (VK n)
        with wfr show ?thesis by(simp add: wf_random_usage'_mono)
      next
        case (SK n)
        with wfr show ?thesis by(simp add: wf_random_usage'_mono)
      next
        case (PAIR t1 t2)
        with ks obtain bl1 bl2 where [simp]: "c_data = [bl1, bl2]" "C = PAIR_C"
          and bl: "bl1 < length f" "bl2 < length f" by auto
        from bl have "f ! bl1 \<in> set f" "f ! bl2 \<in> set f" by(auto simp add: in_set_conv_nth)
        hence "random_usage (f ! bl1) \<le> random_usages f" "random_usage (f ! bl2) \<le> random_usages f"
          by(blast intro: member_le_random_usages)+
        with wfr show ?thesis by(auto elim!: wf_random_usage'_mono)
      next
        case STRING
        then show ?thesis 
        proof cases
          case EMPTY thus ?thesis using wfr by(auto intro: wf_random_usage'_mono)
        next
          case (STRING s b)
          with ks obtain bl where [simp]: "c_data = [bl]" "C = STRING_C b" 
            and bl: "bl < length f" by auto
          from bl have "f ! bl \<in> set f" by(auto simp add: in_set_conv_nth)
          from member_le_random_usages[OF this] wfr
          show ?thesis by(auto elim!: wf_random_usage'_mono)
        qed
      next
        case (N n)
        with wfr show ?thesis by(simp add: wf_random_usage'_mono)
      next
        case garbage
        with ks have False by(auto elim!: \<C>.elims)
        thus ?thesis ..
      next
        case (garbageE t n)
        with ks have False by auto
        thus ?thesis ..
      next
        case (garbageSIG t n)
        with ks have False by auto
        thus ?thesis ..
      qed
    qed simp
  qed
next
  case (Output id o_data succ)
  with step wfr show ?thesis by(auto simp add: gpv.set_map)
next
  case (Input id succ)
  with step wfr show ?thesis by(auto split: if_split_asm dest: gen simp add: gpv.set_map Some_eq_bind_conv)
next
  case (Control id out A succs)
  with step wfr show ?thesis by(auto simp add: gpv.set_map)
next
  case (Probabilistic id p succs)
  with step wfr show ?thesis by(auto)
qed

lemma wf_random_usage_hybrid_step:
  "\<lbrakk> wf_random_usage \<Gamma> f; key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v;
     \<And>t \<sigma>'. Some (t, \<sigma>') \<in> results'_gpv (gen_message sim \<sigma>) \<Longrightarrow> wf_random_usage \<Gamma> (f @ [t]); 
     Some ((s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step' sim ((s,v,f),\<sigma>)) \<rbrakk>
  \<Longrightarrow> wf_random_usage \<Gamma> f'"
by(auto simp add: hybrid_step_def gpv.set_map elim!: wf_random_usage_hybrid_step1 split: option.split_asm)

end