theory CoSP_PKES_Serialisation imports
  CoSP_PKES_Computational
begin

subsubsection \<open>Serialisation\<close>

locale comp_impl_\<beta>_base = comp_impl_base +
  fixes return :: "(bitstring, 'm) return"
  and bind :: "(bitstring, 'm) bind"
  and fail :: "'m fail"
  fixes pnonce :: "security \<Rightarrow> protocol_nonce \<Rightarrow> 'm"
  and "record" :: "t \<Rightarrow> bitstring \<Rightarrow> 'm" -- \<open>records the conversion of the given term into the given bitstring in the state\<close>
  and encryption_key :: "security \<Rightarrow> protocol_nonce \<Rightarrow> 'm"
  and encryption :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> plain \<Rightarrow> protocol_nonce \<Rightarrow> 'm"
  -- \<open>encrypts the given bitstring corresponding to the given term with the given nonces
      as key and randomness, resp., and returns the encrypted message\<close>
  and verification_key :: "security \<Rightarrow> protocol_nonce \<Rightarrow> 'm"
  and signature :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> message \<Rightarrow> protocol_nonce \<Rightarrow> 'm"
begin

context fixes \<eta> :: security begin
adhoc_overloading Monad_Syntax.bind "bind"
adhoc_overloading Monad_Syntax.bind "monad.bind_option fail"

definition encryption_key' :: "t \<Rightarrow> protocol_nonce \<Rightarrow> 'm"
where "encryption_key' t n = encryption_key \<eta> n \<bind> record t"

definition verification_key' :: "t \<Rightarrow> protocol_nonce \<Rightarrow> 'm"
where "verification_key' t n = verification_key \<eta> n \<bind> record t"

primrec \<beta> :: "t \<Rightarrow> 'm"
where
  \<beta>_N:
  "\<beta> (N nonce) = 
  (case nonce of PNonce n \<Rightarrow> pnonce \<eta> n \<bind> record (N nonce)
   | ANonce bs \<Rightarrow> record (N nonce) bs)"
| \<beta>_E:
  "\<beta> (E t1 t2 t3) =
  (case t3 of N (PNonce n) \<Rightarrow>
    (case t1 of EK t1' \<Rightarrow>
      if is_N t1' \<and> is_PNonce (the_N t1') then do {
        let n' = the_PNonce (the_N t1');
        encryption_key' t1 n';
        msg2 \<leftarrow> \<beta> t2;
        msg \<leftarrow> encryption \<eta> n' t2 msg2 n;
        record (E t1 t2 t3) msg
      } else do {
        msg1 \<leftarrow> \<beta> t1;
        msg2 \<leftarrow> \<beta> t2;
        r3 \<leftarrow> pnonce \<eta> n;
        msg \<leftarrow> A_e A \<eta> msg1 msg2 r3;
        record (E t1 t2 t3) msg
      }
    | _ \<Rightarrow> fail)
  | N (ANonce msg) \<Rightarrow>
    (case t1 of EK (N (PNonce n)) \<Rightarrow> (* we omit the check that n has been sampled previously *)
        record (E t1 t2 t3) msg
    | _ \<Rightarrow> fail)
  | _ \<Rightarrow> fail)"
| \<beta>_EK: "\<beta> (EK t) =
  (case t of N (PNonce n) \<Rightarrow> encryption_key' (EK t) n
  | N (ANonce msg) \<Rightarrow> record (EK t) msg
  | _ \<Rightarrow> fail)"
| \<beta>_DK: -- \<open> Why do we need this case at all? Key safety mandates that decryption keys be used only 
     as arguments to decryption nodes, which are destructors. So the nonce cannot appear in terms
     converted by @{term \<beta>}, as @{term \<beta>} is only applied to terms that are sent out. \<close>
  "\<beta> (DK t) =
  (case t of N (PNonce n) \<Rightarrow> do {
      r_n \<leftarrow> pnonce \<eta> n;
      msg \<leftarrow> A_dk A \<eta> r_n;
      record (DK t) msg
    }
  | _ \<Rightarrow> fail)"
| \<beta>_SIG: "\<beta> (SIG t1 t2 t3) =
  (case t3 of N (PNonce n') \<Rightarrow>
    (case t1 of SK (N (PNonce n)) \<Rightarrow> do {
        vk \<leftarrow> verification_key' (VK (N (PNonce n))) n;
        msg \<leftarrow> \<beta> t2;
        msg2 \<leftarrow> signature \<eta> n t2 msg n';
        record (SIG t1 t2 t3) msg2
      }
    | _ \<Rightarrow> fail)
  | N (ANonce msg) \<Rightarrow> (case t1 of SK (N n) \<Rightarrow> record (SIG t1 t2 t3) msg | _ \<Rightarrow> fail)
  | _ \<Rightarrow> fail)"
| \<beta>_VK: "\<beta> (VK t) =
  (case t of N (PNonce n) \<Rightarrow> verification_key' (VK t) n
  | N (ANonce msg) \<Rightarrow> record (VK t) msg
  | _ \<Rightarrow> fail)"
| \<beta>_SK: -- \<open>Same discussion as in the @{const DK} case\<close>
  "\<beta> (SK t) = 
  (case t of N (PNonce n) \<Rightarrow> do {
      r_n \<leftarrow> pnonce \<eta> n;
      msg \<leftarrow> A_sk A \<eta> r_n;
      record (SK t) msg
    }
  | _ \<Rightarrow> fail)"
| \<beta>_PAIR: "\<beta> (PAIR t1 t2) = do {
     msg1 \<leftarrow> \<beta> t1;
     msg2 \<leftarrow> \<beta> t2;
     let msg = A_pair A \<eta> msg1 msg2;
     record (PAIR t1 t2) msg
   }"
| \<beta>_STRING: "\<beta> (STRING b t) = do {
        msg \<leftarrow> \<beta> t;
        msg' \<leftarrow> A_string A \<eta> b msg;
        record (STRING b t) msg'
      }"
| \<beta>_EMPTY: "\<beta> EMPTY = A_empty A \<eta> \<bind> record EMPTY"
| \<beta>_garbage: "\<beta> (garbage t) = (case t of N (ANonce msg) \<Rightarrow> record (garbage t) msg | _ \<Rightarrow> fail)"
| \<beta>_garbageE: "\<beta> (garbageE t1 t2) = (case t2 of N (ANonce msg) \<Rightarrow> record (garbageE t1 t2) msg | _ \<Rightarrow> fail)"
| \<beta>_garbageSIG: "\<beta> (garbageSIG t1 t2) = (case t2 of N (ANonce msg) \<Rightarrow> record (garbageSIG t1 t2) msg | _ \<Rightarrow> fail)"

simps_of_case \<beta>_N_simps: \<beta>_N (splits: t.split nonce.split)
simps_of_case \<beta>_E_simps: \<beta>_E (splits: t.split nonce.split)
simps_of_case \<beta>_EK_simps: \<beta>_EK (splits: t.split nonce.split)
simps_of_case \<beta>_DK_simps: \<beta>_DK (splits: t.split nonce.split)
simps_of_case \<beta>_SIG_simps: \<beta>_SIG (splits: t.split nonce.split)
simps_of_case \<beta>_VK_simps: \<beta>_VK (splits: t.split nonce.split)
simps_of_case \<beta>_SK_simps: \<beta>_SK (splits: t.split nonce.split)
simps_of_case \<beta>_garbage_simps: \<beta>_garbage (splits: t.split nonce.split)
simps_of_case \<beta>_garbageE_simps: \<beta>_garbageE (splits: t.split nonce.split)
simps_of_case \<beta>_garbageSIG_simps: \<beta>_garbageSIG (splits: t.split nonce.split)

lemmas \<beta>_simps =
  \<beta>_N_simps \<beta>_E_simps \<beta>_EK_simps \<beta>_DK_simps \<beta>_SIG_simps \<beta>_VK_simps \<beta>_SK_simps
  \<beta>_PAIR \<beta>_STRING \<beta>_EMPTY
  \<beta>_garbage_simps \<beta>_garbageE_simps \<beta>_garbageSIG_simps

end

end

lemma \<beta>_cong [cong]: -- \<open>The simplifier should not work on the locale parameters\<close>
  "\<And>bind return pnonce.
  comp_impl_\<beta>_base.\<beta> A m pnonce record encryption_key encryption verification_key signature = 
  comp_impl_\<beta>_base.\<beta> A m pnonce record encryption_key encryption verification_key signature"
..

text \<open> @{text \<beta>_context} contains all focuses at which @{term \<beta>} calls itself recursively. \<close> 

(* inductive \<beta>_context :: "t_context \<Rightarrow> t \<Rightarrow> bool"
where
  Top: "\<beta>_context Top t"
| E_EK: "\<lbrakk> \<beta>_context c (E (EK t1) t2 (N (PNonce n))); \<not> is_N t1 \<or> \<not> is_PNonce (the_N t1) \<rbrakk> \<Longrightarrow> \<beta>_context (EKc (Ec1 c t2 (N (PNonce n)))) t1" 
| E_plain: "\<beta>_context c (E t1 t2 (N (PNonce n))) \<Longrightarrow> \<beta>_context (Ec2 t1 c (N (PNonce n))) t2"
| SIG: "\<beta>_context c (SIG (SK (N (PNonce n))) t2 (N (PNonce n'))) \<Longrightarrow> \<beta>_context (SIGc2 (SK (N (PNonce n))) c (N (PNonce n'))) t2"
| PAIR1: "\<beta>_context c (PAIR t1 t2) \<Longrightarrow> \<beta>_context (PAIRc1 c t2) t1"
| PAIR2: "\<beta>_context c (PAIR t1 t2) \<Longrightarrow> \<beta>_context (PAIRc2 t1 c) t2"
| STRING: "\<beta>_context c (STRING b t) \<Longrightarrow> \<beta>_context (STRINGc b c) t"

inductive_simps \<beta>_context_simps [simp]:
  "\<beta>_context Top t"
  "\<beta>_context (Ec1 c t1 t2) t"
  "\<beta>_context (Ec2 t1 c t2) t"
  "\<beta>_context (Ec3 t1 t2 c) t"
  "\<beta>_context (EKc c) t"
  "\<beta>_context (DKc c) t"
  "\<beta>_context (SIGc1 c t2 t3) t"
  "\<beta>_context (SIGc2 t1 c t3) t"
  "\<beta>_context (SIGc3 t1 t2 c) t"
  "\<beta>_context (VKc c) t"
  "\<beta>_context (SKc c) t"
  "\<beta>_context (PAIRc1 c t2) t"
  "\<beta>_context (PAIRc2 t1 c) t"
  "\<beta>_context (STRINGc bool c) t"
  "\<beta>_context (garbagec c) t"
  "\<beta>_context (garbageEc1 c t2) t"
  "\<beta>_context (garbageEc2 t1 c) t"
  "\<beta>_context (garbageSIGc1 c t2) t"
  "\<beta>_context (garbageSIGc2 t1 c) t" *)

text \<open>
  Focuses are not a good abstraction, because we would need preservation assumptions for the
  invariant for all possible movements. So we focus on the subterms on which @{term \<beta>} will be called instead.
\<close>

inductive \<beta>_rec1 :: "t \<Rightarrow> t \<Rightarrow> bool"
where
  E_EK: "\<beta>_rec1 (E (EK t1) t2 (N (PNonce n))) (EK t1)"
| E_plain: "\<beta>_rec1 (E t1 t2 (N (PNonce n))) t2"
| SIG_VK: -- \<open>This one is only needed because @{term \<beta>} calls @{term verification_key} in the case for @{const SIG}.\<close>
  "\<beta>_rec1 (SIG (SK (N (PNonce n))) t (N (PNonce n'))) (VK (N (PNonce n)))"
| SIG: "\<beta>_rec1 (SIG (SK (N (PNonce n))) t2 (N (PNonce n'))) t2"
| PAIR1: "\<beta>_rec1 (PAIR t1 t2) t1"
| PAIR2: "\<beta>_rec1 (PAIR t1 t2) t2"
| STRING: "\<beta>_rec1 (STRING b t) t"

inductive_simps \<beta>_rec1_simps [simp]:
  "\<beta>_rec1 (E t1 t2 t3) t'"
  "\<beta>_rec1 (EK t) t'"
  "\<beta>_rec1 (DK t) t'"
  "\<beta>_rec1 (SIG t1 t2 t3) t'"
  "\<beta>_rec1 (VK t) t'"
  "\<beta>_rec1 (SK t) t'"
  "\<beta>_rec1 (PAIR t1 t2) t'"
  "\<beta>_rec1 EMPTY t'"
  "\<beta>_rec1 (STRING b t) t'"
  "\<beta>_rec1 (N n) t'"
  "\<beta>_rec1 (garbage t) t'"
  "\<beta>_rec1 (garbageE t1 t2) t'"
  "\<beta>_rec1 (garbageSIG t1 t2) t'"

text \<open>We take the transitive (non-reflexive) closure because we only want to have the proper sub-calls\<close>
abbreviation \<beta>_rect :: "t \<Rightarrow> t \<Rightarrow> bool"
where "\<beta>_rect \<equiv> \<beta>_rec1^++"

lemma \<beta>_rect_N [simp]: "\<not> \<beta>_rect (N nonce) t"
by(auto elim: converse_tranclpE)

abbreviation \<beta>_rec :: "t \<Rightarrow> t \<Rightarrow> bool"
where "\<beta>_rec \<equiv> \<beta>_rec1^**"

lemma \<beta>_rec_N [simp]: "\<beta>_rec (N nonce) t \<longleftrightarrow> t = N nonce"
by(auto elim: converse_rtranclpE)

lemma \<beta>_rec_E: "\<beta>_rec (E t1 t2 t3) t \<longleftrightarrow> t = E t1 t2 t3 \<or>
  (\<exists>n. t3 = N (PNonce n) \<and> (\<beta>_rec t2 t \<or> (\<exists>t1'. t1 = EK t1' \<and> \<beta>_rec t1 t)))"
by(auto 4 3 elim: converse_rtranclpE intro: converse_rtranclp_into_rtranclp)

lemma \<beta>_rec_EK [simp]: "\<beta>_rec (EK t) t' \<longleftrightarrow> t' = EK t"
by(auto elim: converse_rtranclpE)

lemma \<beta>_rec_DK [simp]: "\<beta>_rec (DK t) t' \<longleftrightarrow> t' = DK t"
by(auto elim: converse_rtranclpE)

lemma \<beta>_rec_SIG: "\<beta>_rec (SIG t1 t2 t3) t \<longleftrightarrow> t = SIG t1 t2 t3 \<or> 
  (\<exists>n r. t1 = SK (N (PNonce n)) \<and> t3 = N (PNonce r) \<and> (\<beta>_rec t2 t \<or> \<beta>_rec (VK (N (PNonce n))) t))"
by(auto 4 4 elim: converse_rtranclpE intro: converse_rtranclp_into_rtranclp)

lemma \<beta>_rec_VK [simp]: "\<beta>_rec (VK t) t' \<longleftrightarrow> t' = VK t"
by(auto elim: converse_rtranclpE)

lemma \<beta>_rec_SK [simp]: "\<beta>_rec (SK t) t' \<longleftrightarrow> t' = SK t"
by(auto elim: converse_rtranclpE)

lemma \<beta>_rec_PAIR [simp]: "\<beta>_rec (PAIR t1 t2) t \<longleftrightarrow> t = PAIR t1 t2 \<or> \<beta>_rec t1 t \<or> \<beta>_rec t2 t"
by(auto elim: converse_rtranclpE intro: converse_rtranclp_into_rtranclp)

lemma \<beta>_rec_EMPTY [simp]: "\<beta>_rec EMPTY t \<longleftrightarrow> t = EMPTY"
by(auto elim: converse_rtranclpE)

lemma \<beta>_rec_STRING [simp]: "\<beta>_rec (STRING b t) t' \<longleftrightarrow> t' = STRING b t \<or> \<beta>_rec t t'"
by(auto elim: converse_rtranclpE intro: converse_rtranclp_into_rtranclp)

lemma \<beta>_rec_garbage [simp]: "\<beta>_rec (garbage t) t' \<longleftrightarrow> t' = garbage t"
by(auto elim: converse_rtranclpE)

lemma \<beta>_rec_garbageE [simp]: "\<beta>_rec (garbageE t1 t2) t \<longleftrightarrow> t = garbageE t1 t2"
by(auto elim: converse_rtranclpE)

lemma \<beta>_rec_garbageSIG [simp]: "\<beta>_rec (garbageSIG t1 t2) t \<longleftrightarrow> t = garbageSIG t1 t2"
by(auto elim: converse_rtranclpE)


text \<open>@{text \<beta>_nonce} relates terms with the nonces on which @{text pnonce} is called\<close>

inductive \<beta>_nonce :: "t \<Rightarrow> protocol_nonce \<Rightarrow> bool"
where
  N: "\<beta>_nonce (N (PNonce n)) n"
| E_randomness: 
  "\<not> is_EK key \<or> \<not> is_N (the_EK key) \<or> \<not> is_PNonce (the_N (the_EK key))
  \<Longrightarrow> \<beta>_nonce (E key t (N (PNonce n))) n"
| DK: "\<beta>_nonce (DK (N (PNonce n))) n"
| SK: "\<beta>_nonce (SK (N (PNonce n))) n"

inductive_simps \<beta>_nonce_simps [simp]:
  "\<beta>_nonce (E t1 t2 t3) n"
  "\<beta>_nonce (EK t) n"
  "\<beta>_nonce (DK t) n"
  "\<beta>_nonce (SIG t1 t2 t3) n"
  "\<beta>_nonce (VK t) n"
  "\<beta>_nonce (SK t) n"
  "\<beta>_nonce (PAIR t1 t2) n"
  "\<beta>_nonce EMPTY n"
  "\<beta>_nonce (STRING b t) n"
  "\<beta>_nonce (N n') n"
  "\<beta>_nonce (garbage t) n"
  "\<beta>_nonce (garbageE t1 t2) n"
  "\<beta>_nonce (garbageSIG t1 t2) n"

text \<open>@{text \<beta>_nonce1} relates terms with the nonces on which the first simulator calls @{text pnonce}\<close>

inductive \<beta>_nonce1 :: "t \<Rightarrow> protocol_nonce \<Rightarrow> bool"
where
  N: "\<beta>_nonce1 (N (PNonce n)) n"
| E_randomness: -- \<open>The first simulator samples the randomness itself even for keys generated by the protocol\<close>
  "\<beta>_nonce1 (E key t (N (PNonce n))) n"
| EK: "\<beta>_nonce1 (EK (N (PNonce n))) n"
| DK: "\<beta>_nonce1 (DK (N (PNonce n))) n"
| VK: "\<beta>_nonce1 (VK (N (PNonce n))) n"
| SK: "\<beta>_nonce1 (SK (N (PNonce n))) n"
| SIG_randomness: -- \<open>The first simulator samples the randomness itself even for keys generated by the protocol\<close>
  "\<beta>_nonce1 (SIG key t (N (PNonce n))) n"

lemma \<beta>_nonce1_\<beta>_nonce: "\<beta>_nonce t n \<Longrightarrow> \<beta>_nonce1 t n"
by(auto elim: \<beta>_nonce.cases intro: \<beta>_nonce1.intros)

inductive_simps \<beta>_nonce1_simps [simp]:
  "\<beta>_nonce1 (E t1 t2 t3) n"
  "\<beta>_nonce1 (EK t) n"
  "\<beta>_nonce1 (DK t) n"
  "\<beta>_nonce1 (SIG t1 t2 t3) n"
  "\<beta>_nonce1 (VK t) n"
  "\<beta>_nonce1 (SK t) n"
  "\<beta>_nonce1 (PAIR t1 t2) n"
  "\<beta>_nonce1 EMPTY n"
  "\<beta>_nonce1 (STRING b t) n"
  "\<beta>_nonce1 (N n') n"
  "\<beta>_nonce1 (garbage t) n"
  "\<beta>_nonce1 (garbageE t1 t2) n"
  "\<beta>_nonce1 (garbageSIG t1 t2) n"

context fixes \<beta>_nonce :: "t \<Rightarrow> protocol_nonce \<Rightarrow> bool" begin

definition \<beta>_pnonces :: "t \<Rightarrow> protocol_nonce set"
where "\<beta>_pnonces t = {n. \<exists>t'. \<beta>_rec t t' \<and> \<beta>_nonce t' n}"

lemma \<beta>_pnoncesI [intro?]: "\<lbrakk> \<beta>_rec t t'; \<beta>_nonce t' n \<rbrakk> \<Longrightarrow> n \<in> \<beta>_pnonces t"
by(auto simp add: \<beta>_pnonces_def)

lemma \<beta>_pnoncesE [elim?]:
  assumes "n \<in> \<beta>_pnonces t"
  obtains t' where "\<beta>_rec t t'" "\<beta>_nonce t' n"
using assms unfolding \<beta>_pnonces_def by blast

lemma in_\<beta>_pnonces: "n \<in> \<beta>_pnonces t \<longleftrightarrow> (\<exists>t'. \<beta>_rec t t' \<and> \<beta>_nonce t' n)"
unfolding \<beta>_pnonces_def by blast

abbreviation sampled :: "t \<Rightarrow> (protocol_nonce \<rightharpoonup> bitstring) \<Rightarrow> bool"
where "sampled t s \<equiv> \<beta>_pnonces t \<subseteq> dom s"

lemma sampled_\<beta>_recD: "\<lbrakk> sampled t s; \<beta>_rec t t' \<rbrakk> \<Longrightarrow> sampled t' s"
by(auto simp add: \<beta>_pnonces_def intro: rtranclp_trans)

end

abbreviation \<beta>1_pnonces :: "t \<Rightarrow> protocol_nonce set"
where "\<beta>1_pnonces \<equiv> \<beta>_pnonces \<beta>_nonce1"

lemma \<beta>1_pnonces_E_PNonce [simp]:
  "\<beta>1_pnonces (E (EK (N (PNonce n))) t2 (N (PNonce r))) = insert n (insert r (\<beta>1_pnonces t2))"
by(fastforce simp add: \<beta>_pnonces_def \<beta>_rec_E)

lemma \<beta>1_pnonces_E_ANonce [simp]:
  "\<beta>1_pnonces (E (EK (N (ANonce n))) t2 (N (PNonce r))) = insert r (\<beta>1_pnonces t2)"
by(fastforce simp add: \<beta>_pnonces_def \<beta>_rec_E)

lemma \<beta>1_pnonces_E_ANonce1 [simp]:
  "\<forall>n. k \<noteq> N (PNonce n) \<Longrightarrow> \<beta>1_pnonces (E (EK k) t2 (N (PNonce r))) = insert r (\<beta>1_pnonces t2 \<union> \<beta>1_pnonces (EK k))"
by(fastforce simp add: \<beta>_pnonces_def \<beta>_rec_E)

lemma \<beta>1_pnonces_E_ANonce3 [simp]: "\<beta>1_pnonces (E t1 t2 (N (ANonce r))) = {}"
by(auto simp add: \<beta>_pnonces_def \<beta>_rec_E)

lemma \<beta>1_pnonces_EK_PNonce [simp]: "\<beta>1_pnonces (EK (N (PNonce n))) = {n}"
by(fastforce simp add: \<beta>_pnonces_def)

lemma \<beta>1_pnonces_EK_ANonce [simp]: "\<beta>1_pnonces (EK (N (ANonce n))) = {}"
by(fastforce simp add: \<beta>_pnonces_def)

lemma \<beta>1_pnonces_DK [simp]: "\<beta>1_pnonces (DK (N (PNonce n))) = {n}"
by(fastforce simp add: \<beta>_pnonces_def)

lemma \<beta>1_pnonces_SIG_PNonce [simp]:
  "\<beta>1_pnonces (SIG (SK (N (PNonce n))) t2 (N (PNonce r))) = insert n (insert r (\<beta>1_pnonces t2))"
by(fastforce simp add: \<beta>_pnonces_def \<beta>_rec_SIG)

lemma \<beta>1_pnonces_SIG_ANonce3 [simp]:
  "\<beta>1_pnonces (SIG t1 t2 (N (ANonce r))) = {}"
by(fastforce simp add: \<beta>_pnonces_def \<beta>_rec_SIG)

lemma \<beta>1_pnonces_SIG_ANonce1 [simp]: 
  "\<beta>1_pnonces (SIG (SK (N (ANonce x2))) t (N (PNonce n))) = {n}"
by(simp add: \<beta>_pnonces_def \<beta>_rec_SIG)

lemma \<beta>1_pnonces_VK_PNonce [simp]: "\<beta>1_pnonces (VK (N (PNonce n))) = {n}"
by(fastforce simp add: \<beta>_pnonces_def)

lemma \<beta>1_pnonces_VK_ANonce [simp]: "\<beta>1_pnonces (VK (N (ANonce m))) = {}"
by(fastforce simp add: \<beta>_pnonces_def)

lemma \<beta>1_pnonces_SK [simp]: "\<beta>1_pnonces (SK (N (PNonce n))) = {n}"
by(fastforce simp add: \<beta>_pnonces_def)

lemma \<beta>1_pnonces_PAIR [simp]: "\<beta>1_pnonces (PAIR t1 t2) = \<beta>1_pnonces t1 \<union> \<beta>1_pnonces t2"
by(fastforce simp add: \<beta>_pnonces_def)

lemma \<beta>1_pnonces_EMPTY [simp]: "\<beta>1_pnonces EMPTY = {}"
by(fastforce simp add: \<beta>_pnonces_def)

lemma \<beta>1_pnonces_STRING [simp]: "\<beta>1_pnonces (STRING b t) = \<beta>1_pnonces t"
by(fastforce simp add: \<beta>_pnonces_def)

lemma \<beta>1_pnonces_N_PNonce [simp]: "\<beta>1_pnonces (N (PNonce n)) = {n}"
by(fastforce simp add: \<beta>_pnonces_def)

lemma \<beta>1_pnonces_N_ANonce [simp]: "\<beta>1_pnonces (N (ANonce n)) = {}"
by(fastforce simp add: \<beta>_pnonces_def)

lemma \<beta>1_pnonces_garbage [simp]: "\<beta>1_pnonces (garbage t) = {}"
by(fastforce simp add: \<beta>_pnonces_def)

lemma \<beta>1_pnonces_garbageE [simp]: "\<beta>1_pnonces (garbageE t1 t2) = {}"
by(fastforce simp add: \<beta>_pnonces_def)

lemma \<beta>1_pnonces_garbageSIG [simp]: "\<beta>1_pnonces (garbageSIG t1 t2) = {}"
by(fastforce simp add: \<beta>_pnonces_def)

lemma finite_\<beta>1_pnonces [simp]: "finite (\<beta>1_pnonces t)"
proof(rule finite_subset)
  show "\<beta>1_pnonces t \<subseteq> pnonces t"
    by(induction t)(fastforce simp add: \<beta>_rec_E \<beta>_rec_SIG in_\<beta>_pnonces)+
qed simp

lemma \<beta>_induct [case_names E EK DK SIG VK SK PAIR EMPTY STRING N garbage garbageE garbageSIG]: (* unused? *)
  assumes "\<And>t1 t2 t3. \<lbrakk> \<beta>_rec t (E t1 t2 t3);
    \<And>key r. \<lbrakk> t1 = EK key; t3 = N (PNonce r) \<rbrakk> \<Longrightarrow> P t1; \<And>r. t3 = N (PNonce r) \<Longrightarrow> P t2 \<rbrakk>
    \<Longrightarrow> P (E t1 t2 t3)"
  and "\<And>t'. \<beta>_rec t (EK t') \<Longrightarrow> P (EK t')"
  and "\<And>t'. \<beta>_rec t (DK t') \<Longrightarrow> P (DK t')"
  and "\<And>t1 t2 t3. \<lbrakk> \<beta>_rec t (SIG t1 t2 t3);
     \<And>key r. \<lbrakk> t1 = SK (N (PNonce key)); t3 = N (PNonce r) \<rbrakk> \<Longrightarrow> P t2 \<rbrakk>
     \<Longrightarrow> P (SIG t1 t2 t3)"
  and "\<And>t'. \<beta>_rec t (VK t') \<Longrightarrow> P (VK t')"
  and "\<And>t'. \<beta>_rec t (SK t') \<Longrightarrow> P (SK t')"
  and "\<And>t1 t2. \<lbrakk> \<beta>_rec t (PAIR t1 t2); P t1; P t2 \<rbrakk> \<Longrightarrow> P (PAIR t1 t2)"
  and "\<beta>_rec t EMPTY \<Longrightarrow> P EMPTY"
  and "\<And>b t'. \<lbrakk> \<beta>_rec t (STRING b t'); P t' \<rbrakk> \<Longrightarrow> P (STRING b t')"
  and "\<And>n. \<beta>_rec t (N n) \<Longrightarrow> P (N n)"
  and "\<And>t'. \<beta>_rec t (garbage t') \<Longrightarrow> P (garbage t')"
  and "\<And>t1 t2. \<beta>_rec t (garbageE t1 t2) \<Longrightarrow> P (garbageE t1 t2)"
  and "\<And>t1 t2. \<beta>_rec t (garbageSIG t1 t2) \<Longrightarrow> P (garbageSIG t1 t2)"
  shows "P t"
proof -
  def t' == t
  hence "\<beta>_rec t t'" by simp
  thus "P t'"
  proof(induction t' rule: t.induct)
    case (E t1 t2 t3)
    show ?case using E.prems E.IH(1,2)[OF rtranclp.rtrancl_into_rtrancl[OF E.prems]]
      by(rule assms) simp_all
  next
    case SIG
    show ?case using SIG.prems SIG.IH(2)[OF rtranclp.rtrancl_into_rtrancl[OF SIG.prems]]
      by(rule assms)(simp)
  next
    case PAIR
    show ?case using PAIR.prems PAIR.IH[OF rtranclp.rtrancl_into_rtrancl[OF PAIR.prems]]
      by(rule assms) simp_all
  next
    case STRING
    show ?case using STRING.prems STRING.IH[OF rtranclp.rtrancl_into_rtrancl[OF STRING.prems]]
      by(rule assms) simp
  qed(erule assms)+
qed

locale comp_impl_\<beta>_state_monad =
  monad_fail_hom return bind fail return_set bind_set fail_set set
  for A :: implementation
  and return :: "(bitstring \<times> 's, 'm) return"
  and bind :: "(bitstring \<times> 's, 'm) bind"
  and fail :: "'m fail"
  and set :: "'m \<Rightarrow> (bitstring \<times> 's) set"
  and pnonce :: "security \<Rightarrow> protocol_nonce \<Rightarrow> ('s, 'm) stateT"
  and "record" :: "t \<Rightarrow> bitstring \<Rightarrow> ('s, 'm) stateT"
  and encryption_key :: "security \<Rightarrow> protocol_nonce \<Rightarrow> ('s, 'm) stateT"
  and encryption :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> plain \<Rightarrow> protocol_nonce \<Rightarrow> ('s, 'm) stateT"
  and verification_key :: "security \<Rightarrow> protocol_nonce \<Rightarrow> ('s, 'm) stateT"
  and signature :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> message \<Rightarrow> protocol_nonce \<Rightarrow> ('s, 'm) stateT"
begin

declare
  hom_bind [simp]
  hom_return [simp]
  hom_bind_option [simp]

sublocale comp_impl_\<beta>_base
  A
  "return_state return" "bind_state bind" "fail_state fail"
  pnonce "record" encryption_key encryption verification_key signature
by unfold_locales

context notes [intro] = rev_bexI begin

lemma in_set_\<beta>_E:
  "(msg, s''''') \<in> set (run_state (\<beta> \<eta> (E t1 t2 t3)) s) \<longleftrightarrow> 
   (\<exists>n n' s' s'' s''' s'''' m1 m1' m2 m'. t3 = N (PNonce n) \<and> t1 = EK (N (PNonce n')) \<and> 
     (m1', s') \<in> set (run_state (encryption_key \<eta> n') s) \<and>
     (m1, s'') \<in> set (run_state (record (EK (N (PNonce n'))) m1') s') \<and> 
     (m2, s''') \<in> set (run_state (\<beta> \<eta> t2) s'') \<and>
     (m', s'''') \<in> set (run_state (encryption \<eta> n' t2 m2 n) s''') \<and> 
     (msg, s''''') \<in> set (run_state (record (E (EK (N (PNonce n'))) t2 (N (PNonce n))) m') s'''')) \<or>  (\<exists>n t1' s' s'' s''' m1 m2 m' r3. t3 = N (PNonce n) \<and> t1 = EK t1' \<and> (\<forall>n'. t1' \<noteq> N (PNonce n')) \<and>
     (m1, s') \<in> set (run_state (\<beta> \<eta> (EK t1')) s) \<and>
     (m2, s'') \<in> set (run_state (\<beta> \<eta> t2) s') \<and> 
     (r3, s''') \<in> set (run_state (pnonce \<eta> n) s'') \<and> A_e A \<eta> m1 m2 r3 = Some m' \<and>
     (msg, s''''') \<in> set (run_state (record (E (EK t1') t2 (N (PNonce n))) m') s''')) \<or>
  (\<exists>n m'. t3 = N (ANonce m') \<and> t1 = EK (N (PNonce n)) \<and> (msg, s''''') \<in> set (run_state (record (E t1 t2 t3) m') s))"
supply prod.case_cong_weak[cong del]
apply(cases t3; simp)
subgoal for t3'
  apply(cases t3'; simp)
   apply(case_tac [!] t1, simp_all add: encryption_key'_def map_gpv_conv_bind[folded id_def] bind_gpv_assoc Let_def)
   apply(rename_tac [!] t1')
   apply(case_tac [!] t1'; simp)
   subgoal for \<dots> n' by(cases n') fastforce+
  subgoal for \<dots> t1'
    apply(cases t1'; simp)
    subgoal for n' by(cases n')(auto)
    done
  done
done

lemma in_set_\<beta>_EK:
  "(msg, s'') \<in> set (run_state (\<beta> \<eta> (EK t)) s) \<longleftrightarrow>
  (\<exists>n s' m'. t = N (PNonce n) \<and> (m', s') \<in> set (run_state (encryption_key \<eta> n) s) \<and>
    (msg, s'') \<in> set (run_state (record (EK t) m') s')) \<or> 
  (\<exists>msg'. t = N (ANonce msg') \<and> (msg, s'') \<in> set (run_state (record (EK t) msg') s))"
apply(cases t; simp)
apply(rename_tac t')
apply(case_tac t'; auto 4 3 simp add: encryption_key'_def map_gpv_conv_bind[folded id_def] bind_gpv_assoc)
done

lemma in_set_\<beta>_DK:
  "(msg, s'') \<in> set (run_state (\<beta> \<eta> (DK t)) s) \<longleftrightarrow>
  (\<exists>n r_n m' s'. t = N (PNonce n) \<and> (r_n, s') \<in> set (run_state (pnonce \<eta> n) s) \<and> A_dk A \<eta> r_n = Some m' \<and>
    (msg, s'') \<in> set (run_state (record (DK t) m') s'))"
apply(cases t; simp_all)
apply(rename_tac t')
apply(case_tac t'; auto 4 4)
done

lemma in_set_\<beta>_SIG:
  "(msg, s''''') \<in> set (run_state (\<beta> \<eta> (SIG t1 t2 t3)) s) \<longleftrightarrow> 
  (\<exists>n' n s' s'' s''' s'''' vk vk' m2 m'. t3 = N (PNonce n') \<and> t1 = SK (N (PNonce n)) \<and> 
     (vk', s') \<in> set (run_state (verification_key \<eta> n) s) \<and>
     (vk, s'') \<in> set (run_state (record (VK (N (PNonce n))) vk') s') \<and>
     (m2, s''') \<in> set (run_state (\<beta> \<eta> t2) s'') \<and>
     (m', s'''') \<in> set (run_state (signature \<eta> n t2 m2 n') s''') \<and>
     (msg, s''''') \<in> set (run_state (record (SIG (SK (N (PNonce n))) t2 (N (PNonce n'))) m') s'''')) \<or>
  (\<exists>n m'. t3 = N (ANonce m') \<and> t1 = SK (N n) \<and> (msg, s''''') \<in> set (run_state (record (SIG t1 t2 t3) m') s))"apply(cases t3; simp_all)
apply(rename_tac t3')
apply(case_tac t3'; simp_all)
 apply(case_tac [!] t1, simp_all add: verification_key'_def cong del: nonce.case_cong t.case_cong)
 apply(rename_tac [!] t1')
 apply(case_tac [!] t1', simp_all)
apply(rename_tac n')
apply(case_tac n')
apply(auto; fastforce)+
done

lemma in_set_\<beta>_VK:
  "(msg, s'') \<in> set (run_state (\<beta> \<eta> (VK t)) s) \<longleftrightarrow>
  (\<exists>n m' s'. t = N (PNonce n) \<and> (m', s') \<in> set (run_state (verification_key \<eta> n) s) \<and> (msg, s'') \<in> set (run_state (record (VK (N (PNonce n))) m') s')) \<or> 
  (\<exists>m'. t = N (ANonce m') \<and> (msg, s'') \<in> set (run_state (record (VK t) m') s))"
apply(cases t; simp_all)
apply(rename_tac t')
apply(case_tac t'; auto 4 3 simp add: verification_key'_def)
done

lemma in_set_\<beta>_SK:
  "(msg, s'') \<in> set (run_state (\<beta> \<eta> (SK t)) s) \<longleftrightarrow>
  (\<exists>n r_n s' m'. t = N (PNonce n) \<and> (r_n, s') \<in> set (run_state (pnonce \<eta> n) s) \<and> A_sk A \<eta> r_n = Some m' \<and> (msg, s'') \<in> set (run_state (record (SK t) m') s'))"
apply(cases t; simp_all)
apply(rename_tac t')
apply(case_tac t'; auto 4 4)
done

lemma in_set_\<beta>_PAIR:
  "(msg, s''') \<in> set (run_state (\<beta> \<eta> (PAIR t1 t2)) s) \<longleftrightarrow>
  (\<exists>m1 s' m2 m' s''. (m1, s') \<in> set (run_state (\<beta> \<eta> t1) s) \<and> (m2, s'') \<in> set (run_state (\<beta> \<eta> t2) s') \<and> m' = A_pair A \<eta> m1 m2 \<and> (msg, s''') \<in> set (run_state (record (PAIR t1 t2) m') s''))"
by(fastforce simp add: Let_def)

lemma in_set_\<beta>_EMPTY:
  "(msg, s') \<in> set (run_state (\<beta> \<eta> EMPTY) s) \<longleftrightarrow> (\<exists>m'. A_empty A \<eta> = Some m' \<and> (msg, s') \<in> set (run_state (record EMPTY m') s))"
by auto

lemma in_set_\<beta>_STRING:
  "(msg, s'') \<in> set (run_state (\<beta> \<eta> (STRING b t)) s) \<longleftrightarrow>
  (\<exists>m1 s' m'. (m1, s') \<in> set (run_state (\<beta> \<eta> t) s) \<and> A_string A \<eta> b m1 = Some m' \<and> (msg, s'') \<in> set (run_state (record (STRING b t) m') s'))"
by fastforce

lemma in_set_\<beta>_N:
  "(msg, s'') \<in> set (run_state (\<beta> \<eta> (N n)) s) \<longleftrightarrow>
  (\<exists>n' s' m'. n = PNonce n' \<and> (m', s') \<in> set (run_state (pnonce \<eta> n') s) \<and> (msg, s'') \<in> set (run_state (record (N n) m') s')) \<or> 
  (\<exists>m'. n = ANonce m' \<and> (msg, s'') \<in> set (run_state (record (N n) m') s))"
by(cases n) auto

lemma in_set_\<beta>_garbage:
  "(msg, s') \<in> set (run_state (\<beta> \<eta> (garbage t)) s) \<longleftrightarrow>
   (\<exists>m'. t = N (ANonce m') \<and> (msg, s') \<in> set (run_state (record (garbage t) m') s))"
apply(cases t; simp_all)
apply(rename_tac t')
apply(case_tac t'; auto)
done

lemma in_set_\<beta>_garbageE:
  "(msg, s') \<in> set (run_state (\<beta> \<eta> (garbageE t1 t2)) s) \<longleftrightarrow>
  (\<exists>m'. t2 = N (ANonce m') \<and> (msg, s') \<in> set (run_state (record (garbageE t1 t2) m') s))"
apply(cases t2; simp_all)
apply(rename_tac t')
apply(case_tac t'; auto)
done

lemma in_set_\<beta>_garbageSIG:
  "(msg, s') \<in> set (run_state (\<beta> \<eta> (garbageSIG t1 t2)) s) \<longleftrightarrow>
  (\<exists>m'. t2 = N (ANonce m') \<and> (msg, s') \<in> set (run_state (record (garbageSIG t1 t2) m') s))"
apply(cases t2; simp_all)
apply(rename_tac t')
apply(case_tac t'; auto)
done

lemmas in_set_\<beta> = 
  in_set_\<beta>_E in_set_\<beta>_EK in_set_\<beta>_DK in_set_\<beta>_SIG in_set_\<beta>_VK in_set_\<beta>_SK
  in_set_\<beta>_PAIR in_set_\<beta>_EMPTY in_set_\<beta>_STRING in_set_\<beta>_N
  in_set_\<beta>_garbage in_set_\<beta>_garbageE in_set_\<beta>_garbageSIG

end

lemma \<beta>_set_induct [consumes 1, case_names
  E E_key E_adversary EK_PNonce EK_ANonce DK SIG SIG_ANonce VK_PNonce VK_ANonce SK
  PAIR EMPTY STRING PNonce ANonce garbage garbageE garbageSIG]:
  assumes major: "(msg, s') \<in> set (run_state (\<beta> \<eta> t) s)"
  and E: "\<And>k t r s m1 s' m1' s'' m2 s''' m' s'''' msg s'''''.
    \<lbrakk> (m1', s') \<in> set (run_state (encryption_key \<eta> k) s); 
      (m1, s'') \<in> set (run_state (record (EK (N (PNonce k))) m1') s');
      (m2, s''') \<in> set (run_state (\<beta> \<eta> t) s'');
      P t m2 s'' s''';
      (m', s'''') \<in> set (run_state (encryption \<eta> k t m2 r) s''');
      (msg, s''''') \<in> set (run_state (record (E (EK (N (PNonce k))) t (N (PNonce r))) m') s'''') \<rbrakk>
    \<Longrightarrow> P (E (EK (N (PNonce k))) t (N (PNonce r))) msg s s'''''"
  and E_key: "\<And>k t r s m1 s' m2 s'' r3 s''' m' msg s''''.
    \<lbrakk> \<forall>n. k \<noteq> N (PNonce n); 
      (m1, s') \<in> set (run_state (\<beta> \<eta> (EK k)) s);
      P (EK k) m1 s s';
      (m2, s'') \<in> set (run_state (\<beta> \<eta> t) s');
      P t m2 s' s'';
      (r3, s''') \<in> set (run_state (pnonce \<eta> r) s'');
      A_e A \<eta> m1 m2 r3 = Some m'; 
      (msg, s'''') \<in> set (run_state (record (E (EK k) t (N (PNonce r))) m') s''') \<rbrakk>
    \<Longrightarrow> P (E (EK k) t (N (PNonce r))) msg s s''''"
  and E_adversary: "\<And>k t m' s msg s'.
    (msg, s') \<in> set (run_state (record (E (EK (N (PNonce k))) t (N (ANonce m'))) m') s)
    \<Longrightarrow> P (E (EK (N (PNonce k))) t (N (ANonce m'))) msg s s'"
  and EK_PNonce: "\<And>k s m' s' msg s''.
    \<lbrakk> (m', s') \<in> set (run_state (encryption_key \<eta> k) s);
      (msg, s'') \<in> set (run_state (record (EK (N (PNonce k))) m') s') \<rbrakk>
    \<Longrightarrow> P (EK (N (PNonce k))) msg s s''"
  and EK_ANonce: "\<And>msg' s msg s'.
    (msg, s') \<in> set (run_state (record (EK (N (ANonce msg'))) msg') s)
    \<Longrightarrow> P (EK (N (ANonce msg'))) msg s s'"
  and DK: "\<And>n s r_n s' m' msg s''.
    \<lbrakk> (r_n, s') \<in> set (run_state (pnonce \<eta> n) s);
      A_dk A \<eta> r_n = Some m';
      (msg, s'') \<in> set (run_state (record (DK (N (PNonce n))) m') s') \<rbrakk>
    \<Longrightarrow> P (DK (N (PNonce n))) msg s s''"
  and SIG: "\<And>k t r s vk' s' vk s'' m2 s''' m' s'''' msg s'''''.
    \<lbrakk> (vk', s') \<in> set (run_state (verification_key \<eta> k) s);
      (vk, s'') \<in> set (run_state (record (VK (N (PNonce k))) vk') s');
      (m2, s''') \<in> set (run_state (\<beta> \<eta> t) s'');
      P t m2 s'' s''';
      (m', s'''') \<in> set (run_state (signature \<eta> k t m2 r) s'''); 
      (msg, s''''') \<in> set (run_state (record (SIG (SK (N (PNonce k))) t (N (PNonce r))) m') s'''') \<rbrakk>
    \<Longrightarrow> P (SIG (SK (N (PNonce k))) t (N (PNonce r))) msg s s'''''"
  and SIG_ANonce: "\<And>k t m' s msg s'.
    (msg, s') \<in> set (run_state (record (SIG (SK (N k)) t (N (ANonce m'))) m') s)
    \<Longrightarrow> P (SIG (SK (N k)) t (N (ANonce m'))) msg s s'"
  and VK_PNonce: "\<And>n s m' s' msg s''.
    \<lbrakk> (m', s') \<in> set (run_state (verification_key \<eta> n) s);
      (msg, s'') \<in> set (run_state (record (VK (N (PNonce n))) m') s') \<rbrakk>
    \<Longrightarrow> P (VK (N (PNonce n))) msg s s''"
  and VK_ANonce: "\<And>m s msg s'.
    (msg, s') \<in> set (run_state (record (VK (N (ANonce m))) m) s)
    \<Longrightarrow> P (VK (N (ANonce m))) msg s s'"
  and SK: "\<And>n s r_n s' m' msg s''.
    \<lbrakk> (r_n, s') \<in> set (run_state (pnonce \<eta> n) s); 
      A_sk A \<eta> r_n = Some m';
      (msg, s'') \<in> set (run_state (record (SK (N (PNonce n))) m') s')
    \<rbrakk>
    \<Longrightarrow> P (SK (N (PNonce n))) msg s s''"
  and PAIR: "\<And>t1 t2 s m1 s' m2 s'' m' msg s'''.
    \<lbrakk> (m1, s') \<in> set (run_state (\<beta> \<eta> t1) s);
      P t1 m1 s s';
      (m2, s'') \<in> set (run_state (\<beta> \<eta> t2) s');
      P t2 m2 s' s'';
      m' = A_pair A \<eta> m1 m2; 
      (msg, s''') \<in> set (run_state (record (PAIR t1 t2) m') s'')
    \<rbrakk>
    \<Longrightarrow> P (PAIR t1 t2) msg s s'''"
  and EMPTY: "\<And>s m' msg s'.
    \<lbrakk> A_empty A \<eta> = Some m'; (msg, s') \<in> set (run_state (record EMPTY m') s) \<rbrakk>
    \<Longrightarrow> P EMPTY msg s s'"
  and STRING: "\<And>b t s m1 s' m' msg s''.
    \<lbrakk> (m1, s') \<in> set (run_state (\<beta> \<eta> t) s); 
      P t m1 s s';
      A_string A \<eta> b m1 = Some m';
      (msg, s'') \<in> set (run_state (record (STRING b t) m') s') \<rbrakk>
    \<Longrightarrow> P (STRING b t) msg s s''"
  and PNonce: "\<And>n s m' s' msg s''.
    \<lbrakk> (m', s') \<in> set (run_state (pnonce \<eta> n) s); 
      (msg, s'') \<in> set (run_state (record (N (PNonce n)) m') s') \<rbrakk>
    \<Longrightarrow> P (N (PNonce n)) msg s s''"
  and ANonce: "\<And>m s msg s'.
    (msg, s') \<in> set (run_state (record (N (ANonce m)) m) s)
    \<Longrightarrow> P (N (ANonce m)) msg s s'"
  and garbage: "\<And>m s msg s'.
    (msg, s') \<in> set (run_state (record (garbage (N (ANonce m))) m) s)
    \<Longrightarrow> P (garbage (N (ANonce m))) msg s s'"
  and garbageE: "\<And>t m s msg s'.
    (msg, s') \<in> set (run_state (record (garbageE t (N (ANonce m))) m) s)
    \<Longrightarrow> P (garbageE t (N (ANonce m))) msg s s'"
  and garbageSIG: "\<And>t m s msg s'.
    (msg, s') \<in> set (run_state (record (garbageSIG t (N (ANonce m))) m) s)
    \<Longrightarrow> P (garbageSIG t (N (ANonce m))) msg s s'"
  notes \<beta>.simps[simp del] in_set_\<beta>[simp]
  shows "P t msg s s'"
using major
proof(induction t arbitrary: s msg s')
  { case *: E          then show ?case by(auto intro: E E_key E_adversary) }
  { case EK            then show ?case by(auto intro: EK_ANonce EK_PNonce) }
  { case *: DK         then show ?case by(auto intro: DK) }
  { case *: SIG        then show ?case by(auto intro: SIG SIG_ANonce) }
  { case VK            then show ?case by(auto intro: VK_ANonce VK_PNonce) }
  { case *: SK         then show ?case by(auto intro: SK) }
  { case *: PAIR       then show ?case by(auto intro: PAIR) }
  { case *: EMPTY      then show ?case by(auto intro: EMPTY) }
  { case *: STRING     then show ?case by(auto intro: STRING) }
  { case N             then show ?case by(auto intro: ANonce PNonce) }
  { case *: garbage    then show ?case by(auto intro: garbage) }
  { case *: garbageE   then show ?case by(auto intro: garbageE) }
  { case *: garbageSIG then show ?case by(auto intro: garbageSIG) }
qed

end

context includes lifting_syntax begin

lemma encryption_key'_parametric [transfer_rule]:
  "((M ===> (op = ===> M) ===> M)
   ===> (op = ===> op = ===> M)
   ===> (op = ===> op = ===> M)
   ===> op = ===> op = ===> op = ===> M)
   comp_impl_\<beta>_base.encryption_key' comp_impl_\<beta>_base.encryption_key'"
unfolding comp_impl_\<beta>_base.encryption_key'_def[abs_def] by transfer_prover

lemma verification_key'_parametric [transfer_rule]:
  "((M ===> (op = ===> M) ===> M)
   ===> (op = ===> op = ===> M)
   ===> (op = ===> op = ===> M)
   ===> op = ===> op = ===> op = ===> M)
   comp_impl_\<beta>_base.verification_key' comp_impl_\<beta>_base.verification_key'"
unfolding comp_impl_\<beta>_base.verification_key'_def[abs_def] by transfer_prover

term comp_impl_\<beta>_base.\<beta>

lemma \<beta>_parametric [transfer_rule]:
  fixes M :: "'m1 \<Rightarrow> 'm2 \<Rightarrow> bool" and X :: "'more1 \<Rightarrow> 'more2 \<Rightarrow> bool" shows
  "(op = 
   ===> (M ===> (op = ===> M) ===> M) ===> M
   ===> (op = ===> op = ===> M)
   ===> (op = ===> op = ===> M)
   ===> (op = ===> op = ===> M)
   ===> (op = ===> op = ===> op = ===> op = ===> op = ===> M)
   ===> (op = ===> op = ===> M)
   ===> (op = ===> op = ===> op = ===> op = ===> op = ===> M)
   ===> op = ===> op = ===> M) comp_impl_\<beta>_base.\<beta> comp_impl_\<beta>_base.\<beta>"
proof(rule rel_funI; rule rel_funI; rule rel_funI; rule rel_funI; rule rel_funI; rule rel_funI; rule rel_funI; rule rel_funI; rule rel_funI; rule rel_funI; hypsubst)
  fix A :: implementation
    and bind1 :: "(bitstring, 'm1) bind"
    and fail1 :: "'m1 fail"
    and pnonce1 :: "security \<Rightarrow> protocol_nonce \<Rightarrow> 'm1"
    and record1 :: "t \<Rightarrow> bitstring \<Rightarrow> 'm1"
    and encryption_key1 :: "security \<Rightarrow> protocol_nonce \<Rightarrow> 'm1"
    and encryption1 :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> plain \<Rightarrow> protocol_nonce \<Rightarrow> 'm1"
    and verification_key1 :: "security \<Rightarrow> protocol_nonce \<Rightarrow> 'm1"
    and signature1 :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> message \<Rightarrow> protocol_nonce \<Rightarrow> 'm1"
    and bind2 :: "(bitstring, 'm2) bind"
    and fail2 :: "'m2 fail"
    and pnonce2 :: "security \<Rightarrow> protocol_nonce \<Rightarrow> 'm2"
    and record2 :: "t \<Rightarrow> bitstring \<Rightarrow> 'm2"
    and encryption_key2 :: "security \<Rightarrow> protocol_nonce \<Rightarrow> 'm2"
    and encryption2 :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> plain \<Rightarrow> protocol_nonce \<Rightarrow> 'm2"
    and verification_key2 :: "security \<Rightarrow> protocol_nonce \<Rightarrow> 'm2"
    and signature2 :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> message \<Rightarrow> protocol_nonce \<Rightarrow> 'm2"
    and \<eta> :: security
  assume bind: "(M ===> (op = ===> M) ===> M) bind1 bind2"
    and fail: "M fail1 fail2"
    and pnonce: "(op = ===> op = ===> M) pnonce1 pnonce2"
    and "record": "(op = ===> op = ===> M) record1 record2"
    and encryption_key: "(op = ===> op = ===> M) encryption_key1 encryption_key2"
    and "(op = ===> op = ===> op = ===> op = ===> op = ===> M) encryption1 encryption2"
    and verification_key: "(op = ===> op = ===> M) verification_key1 verification_key2"
    and signature: "(op = ===> op = ===> op = ===> op = ===> op = ===> M) signature1 signature2"
  note [transfer_rule] = this[folded list.rel_eq]
  show "(op = ===> M)
             (comp_impl_\<beta>_base.\<beta> A bind1 fail1 pnonce1 record1 encryption_key1 encryption1 verification_key1 signature1 \<eta>)
             (comp_impl_\<beta>_base.\<beta> A bind2 fail2 pnonce2 record2 encryption_key2 encryption2 verification_key2 signature2 \<eta>)"
    unfolding comp_impl_\<beta>_base.\<beta>_def[abs_def] comp_impl_\<beta>_base.encryption_key'_def comp_impl_\<beta>_base.verification_key'_def map_def
    apply(rule t.rec_transfer[THEN rel_funD, THEN rel_funD, THEN rel_funD, THEN rel_funD, THEN rel_funD, THEN rel_funD, THEN rel_funD, THEN rel_funD, THEN rel_funD, THEN rel_funD, THEN rel_funD, THEN rel_funD, THEN rel_funD, where ?A5="M"])
    apply transfer_prover+
    done
qed

end

context comp_impl_\<beta>_state_monad begin
lemma \<beta>_invariant [consumes 1, case_names pnonce "record" encryption_key encryption verification_key signature]:
  assumes "(msg, s') \<in> set (run_state (\<beta> \<eta> t) s)"
  and pnonce: "\<And>s msg s' t' n. \<lbrakk> (msg, s') \<in> set (run_state (pnonce \<eta> n) s); \<beta>_rec t t'; \<beta>_nonce t' n \<rbrakk> \<Longrightarrow> I s s'"
  and "record": "\<And>s t' msg msg' s'. \<lbrakk> (msg', s') \<in> set (run_state (record t' msg) s); \<beta>_rec t t' \<rbrakk> \<Longrightarrow> I s s'" (* FIXME: restrict m to correctly serialised terms *)
  and encryption_key:
    "\<And>s n msg s' msg' s''. \<lbrakk> (msg, s') \<in> set (run_state (encryption_key \<eta> n) s); \<beta>_rec t (EK (N (PNonce n)));
       (msg', s'') \<in> set (run_state (record (EK (N (PNonce n))) msg) s') \<rbrakk>
    \<Longrightarrow> I s s''"
  and encryption:
   "\<And>s n t' bs r msg s' msg' s''. \<lbrakk> (msg, s') \<in> set (run_state (encryption \<eta> n t' bs r) s); \<beta>_rec t (E (EK (N (PNonce n))) t' (N (PNonce r)));
      (msg', s'') \<in> set (run_state (record (E (EK (N (PNonce n))) t' (N (PNonce r))) msg) s') \<rbrakk>
    \<Longrightarrow> I s s''" (* FIXME: restrict bs to correctly serialised terms *)
  and verification_key:
    "\<And>s n msg s' msg' s''. \<lbrakk> (msg, s') \<in> set (run_state (verification_key \<eta> n) s); \<beta>_rec t (VK (N (PNonce n)));
      (msg', s'') \<in> set (run_state (record (VK (N (PNonce n))) msg) s') \<rbrakk>
    \<Longrightarrow> I s s''"
  and signature: "\<And>s n t' bs r msg s' msg' s''. \<lbrakk> (msg, s') \<in> set (run_state (signature \<eta> n t' bs r) s); \<beta>_rec t (SIG (SK (N (PNonce n))) t' (N (PNonce r)));
      (msg', s'') \<in> set (run_state (record (SIG (SK (N (PNonce n))) t' (N (PNonce r))) msg) s') \<rbrakk>
    \<Longrightarrow> I s s''" (* FIXME: restrict bs to correctly serialised terms *)
  shows "I^++ s s'"
proof -
  def t' \<equiv> t
  with assms(1) have "(msg, s') \<in> set (run_state (\<beta> \<eta> t') s)" "\<beta>_rec t t'" by simp_all
  then show ?thesis
  proof(induction t' msg s s' rule: \<beta>_set_induct)
    case (E k t' r s m1 s' m1' s'' m2 s''' m' s'''' msg s''''')
    have "\<beta>_rec t (EK (N (PNonce k)))" using E.prems by(auto elim!: rtranclp_trans)
    with E.hyps(1) have "I s s''" using E.hyps(2) by(rule encryption_key) 
    also from E.prems have "\<beta>_rec t t'" by(auto elim!: rtranclp_trans)
    hence "I\<^sup>+\<^sup>+ s'' s'''" by(rule E.IH)
    also have "I s''' s'''''" using E.hyps(4) E.prems E.hyps(5) by(rule encryption)
    finally show ?case .
  next
    case (E_key k t' r s m1 s' m2 s'' r3 s''' m' msg s'''')
    from E_key.prems E_key.hyps(1) have "\<beta>_rec t (EK k)"
      by(cases k)(auto elim!: rtranclp_trans simp add: is_PNonce_def)
    then have "I\<^sup>+\<^sup>+ s s'" by(rule E_key.IH)
    also from E_key.prems have "\<beta>_rec t t'" by(auto elim!: rtranclp_trans)
    then have "I\<^sup>+\<^sup>+ s' s''" by(rule E_key.IH)
    also have "\<beta>_nonce (E (EK k) t' (N (PNonce r))) r"
      using E_key.hyps(1) by(cases k)(simp_all add: is_PNonce_def)
    with E_key.hyps(4) E_key.prems have "I s'' s'''" by(rule pnonce)
    also have "I s''' s''''" using E_key.hyps(6) E_key.prems by(rule "record")
    finally show ?case .
  next
    case (E_adversary k t' m' s msg s')
    then have "I s s'" by(rule "record")
    thus ?case ..
  next
    case (EK_PNonce k s m' s' msg s'')
    from EK_PNonce(1,3,2) have "I s s''" by(rule encryption_key)
    thus ?case ..
  next
    case (EK_ANonce msg' s msg s')
    then have "I s s'" by(rule "record")
    thus ?case ..
  next
    case (DK n s r_n s' m' msg s'')
    from DK.hyps(1) DK.prems have "I s s'" by(rule pnonce) simp
    moreover from DK.hyps(3) DK.prems have "I s' s''" by(rule "record")
    ultimately show ?case by(rule r_r_into_tranclp)
  next
    case (SIG k t' r s vk' s' vk s'' m2 s''' m' s'''' msg s''''')
    from SIG.prems have "\<beta>_rec t (VK (N (PNonce k)))" by(auto elim!: rtranclp_trans)
    with SIG.hyps(1) have "I s s''" using SIG.hyps(2) by(rule verification_key)
    also have "\<beta>_rec t t'" using SIG.prems by(auto elim!: rtranclp_trans)
    then have "I\<^sup>+\<^sup>+ s'' s'''" by(rule SIG.IH)
    also have "I s''' s'''''" using SIG.hyps(4) SIG.prems SIG.hyps(5) by(rule signature)
    finally show ?case .
  next
    case (SIG_ANonce k t' m' s msg s')
    then have "I s s'" by(rule "record")
    thus ?case ..
  next
    case (VK_PNonce n s m' s' msg s'')
    have "I s s''" using VK_PNonce(1,3,2) by(rule verification_key)
    thus ?case ..
  next
    case (VK_ANonce m s msg s')
    then have "I s s'" by(rule "record")
    thus ?case ..
  next
    case (SK n s r_n s' m' msg s'')
    have "I s s'" using SK.hyps(1) SK.prems by(rule pnonce) simp
    moreover have "I s' s''" using SK.hyps(3) SK.prems by(rule "record")
    ultimately show ?case by(rule r_r_into_tranclp)
  next
    case (PAIR t1 t2 s m1 s' m2 s'' m' msg s''')
    have "\<beta>_rec t t1" using PAIR.prems by(auto elim!: rtranclp_trans)
    then have "I\<^sup>+\<^sup>+ s s'" by(rule PAIR.IH)
    also have "\<beta>_rec t t2" using PAIR.prems by(auto elim!: rtranclp_trans)
    then have "I\<^sup>+\<^sup>+ s' s''" by(rule PAIR.IH)
    also have "I s'' s'''" using PAIR.hyps(4) PAIR.prems by(rule "record")
    finally show ?case .
  next
    case (EMPTY s m' msg s')
    have "I s s'" using EMPTY.hyps(2) EMPTY.prems by(rule "record")
    thus ?case ..
  next
    case (STRING b t' s m1 s' m' msg s'')
    have "\<beta>_rec t t'" using STRING.prems by(auto elim!: rtranclp_trans)
    then have "I\<^sup>+\<^sup>+ s s'" by(rule STRING.IH)
    also have "I s' s''" using STRING.hyps(3) STRING.prems by(rule "record")
    finally show ?case .
  next
    case (PNonce n s m' s' msg s'')
    have "I s s'" using PNonce.hyps(1) PNonce.prems by(rule pnonce) simp
    moreover have "I s' s''" using PNonce.hyps(2) PNonce.prems by(rule "record")
    ultimately show ?case by(rule r_r_into_tranclp)
  next
    case (ANonce m s msg s')
    hence "I s s'" by(rule "record")
    thus ?case ..
  next
    case (garbage m s msg s')
    hence "I s s'" by(rule "record")
    thus ?case ..
  next
    case (garbageE t' m s msg s')
    hence "I s s'" by(rule "record")
    thus ?case ..
  next
    case (garbageSIG t' m s msg s')
    hence "I s s'" by(rule "record")
    thus ?case ..
  qed
qed

lemma \<beta>_invariant_p [consumes 1, case_names pnonce "record" encryption_key encryption verification_key signature transp]:
  assumes "(msg, s') \<in> set (run_state (\<beta> \<eta> t) s)"
  and pnonce: "\<And>s msg s' t' n. \<lbrakk> (msg, s') \<in> set (run_state (pnonce \<eta> n) s); \<beta>_rec t t'; \<beta>_nonce t' n \<rbrakk> \<Longrightarrow> I s s'"
  and "record": "\<And>s t' msg msg' s'. \<lbrakk> (msg', s') \<in> set (run_state (record t' msg) s); \<beta>_rec t t' \<rbrakk> \<Longrightarrow> I s s'" (* FIXME: restrict m to correctly serialised terms *)
  and encryption_key:
    "\<And>s n msg s' msg' s''. \<lbrakk> (msg, s') \<in> set (run_state (encryption_key \<eta> n) s); \<beta>_rec t (EK (N (PNonce n)));
       (msg', s'') \<in> set (run_state (record (EK (N (PNonce n))) msg) s') \<rbrakk>
    \<Longrightarrow> I s s''"
  and encryption:
   "\<And>s n t' bs r msg s' msg' s''. \<lbrakk> (msg, s') \<in> set (run_state (encryption \<eta> n t' bs r) s); \<beta>_rec t (E (EK (N (PNonce n))) t' (N (PNonce r)));
      (msg', s'') \<in> set (run_state (record (E (EK (N (PNonce n))) t' (N (PNonce r))) msg) s') \<rbrakk>
    \<Longrightarrow> I s s''" (* FIXME: restrict bs to correctly serialised terms *)
  and verification_key:
    "\<And>s n msg s' msg' s''. \<lbrakk> (msg, s') \<in> set (run_state (verification_key \<eta> n) s); \<beta>_rec t (VK (N (PNonce n)));
      (msg', s'') \<in> set (run_state (record (VK (N (PNonce n))) msg) s') \<rbrakk>
    \<Longrightarrow> I s s''"
  and signature: "\<And>s n t' bs r msg s' msg' s''. \<lbrakk> (msg, s') \<in> set (run_state (signature \<eta> n t' bs r) s); \<beta>_rec t (SIG (SK (N (PNonce n))) t' (N (PNonce r)));
      (msg', s'') \<in> set (run_state (record (SIG (SK (N (PNonce n))) t' (N (PNonce r))) msg) s') \<rbrakk>
    \<Longrightarrow> I s s''" (* FIXME: restrict bs to correctly serialised terms *)
  and "transp I"
  shows "I s s'"
apply(subst transp_tranclp_id[OF \<open>transp I\<close>, symmetric])
using assms(1-7) by(rule \<beta>_invariant)

end

text \<open>All the protocol nonces in a term which can be reached by calling @{text \<beta>} on any accessible subterm\<close>

definition \<beta>1_anonces :: "t \<Rightarrow> protocol_nonce set"
where "\<beta>1_anonces t = {n. \<exists>t'. acc_subterm t t' \<and> n \<in> \<beta>1_pnonces t'}"

lemma \<beta>1_anonces_simps [simp]:
  shows \<beta>1_anonces_PNonce: "\<beta>1_anonces (N (PNonce n)) = {n}"
  and \<beta>1_anonces_ANonce: "\<beta>1_anonces (N (ANonce m)) = {}"
  and \<beta>1_anonces_EK: "\<beta>1_anonces (EK k) = \<beta>1_pnonces (EK k)"
  and \<beta>1_anonces_E:
    "\<beta>1_anonces (E (EK k) t r) = \<beta>1_pnonces (EK k) \<union> \<beta>1_pnonces (E (EK k) t r) \<union> \<beta>1_anonces t"
  and \<beta>1_anonces_DK: "\<beta>1_anonces (DK k) = \<beta>1_pnonces (DK k)"
  and \<beta>1_anonces_VK: "\<beta>1_anonces (VK k) = \<beta>1_pnonces (VK k)"
  and \<beta>1_anonces_SK: "\<beta>1_anonces (SK k) = \<beta>1_pnonces (SK k)"
  and \<beta>1_anonces_SIG:
    "\<beta>1_anonces (SIG (SK k) t r) = \<beta>1_pnonces (VK k) \<union> \<beta>1_pnonces (SIG (SK k) t r) \<union> \<beta>1_anonces t"
  and \<beta>1_anonces_PAIR: "\<beta>1_anonces (PAIR t1 t2) = \<beta>1_anonces t1 \<union> \<beta>1_anonces t2"
  and \<beta>1_anonces_EMPTY: "\<beta>1_anonces EMPTY = {}"
  and \<beta>1_anonces_STRING: "\<beta>1_anonces (STRING b t) = \<beta>1_anonces t"
  and \<beta>1_anonces_garbage: "\<beta>1_anonces (garbage t) = {}"
  and \<beta>1_anonces_garbageE: "\<beta>1_anonces (garbageE k t) = \<beta>1_anonces k"
  and \<beta>1_anonces_garbageSIG: "\<beta>1_anonces (garbageSIG k t) = \<beta>1_anonces k"
by(auto simp add: \<beta>1_anonces_def)

lemma \<beta>1_pnonces_le_anonces: "\<beta>1_pnonces t \<subseteq> \<beta>1_anonces t"
by(auto simp add: \<beta>1_anonces_def)

lemma \<beta>1_pnonces_into_anonces: "n \<in> \<beta>1_pnonces t \<Longrightarrow> n \<in> \<beta>1_anonces t"
using \<beta>1_pnonces_le_anonces[of t] by blast

locale comp_impl_\<beta> =
  comp_impl_\<beta>_state_monad A return bind fail set pnonce "record" encryption_key encryption verification_key signature
  + comp_impl A
  for A ::  implementation
  and return :: "(bitstring \<times> 's, 'm) return"
  and bind :: "(bitstring \<times> 's, 'm) bind"
  and fail :: "'m fail"
  and set :: "'m \<Rightarrow> (bitstring \<times> 's) set"
  and pnonce :: "security \<Rightarrow> protocol_nonce \<Rightarrow> ('s, 'm) stateT"
  and "record" :: "t \<Rightarrow> bitstring \<Rightarrow> ('s, 'm) stateT"
  and encryption_key :: "security \<Rightarrow> protocol_nonce \<Rightarrow> ('s, 'm) stateT"
  and encryption :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> plain \<Rightarrow> protocol_nonce \<Rightarrow> ('s, 'm) stateT"
  and verification_key :: "security \<Rightarrow> protocol_nonce \<Rightarrow> ('s, 'm) stateT"
  and signature :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> message \<Rightarrow> protocol_nonce \<Rightarrow> ('s, 'm) stateT"
begin

lemma len_of_\<beta>:
  notes [simp del] = \<beta>.simps and [simp] = \<beta>_simps
  assumes \<beta>: "(msg, s') \<in> set (run_state (\<beta> \<eta> t) s)"
  and pnonce: "\<And>n s. \<forall>(r_n, _)\<in>set (run_state (pnonce \<eta> n) s). length r_n = len_of_P_nonce \<eta>"
  and "record" [dest!]: "\<And>m m' t s s'. (m', s') \<in> set (run_state (record t m) s) \<Longrightarrow> length m' = length m"
  and encryption_key: "\<And>n s m s'. (m, s') \<in> set (run_state (encryption_key \<eta> n) s) \<Longrightarrow> length m = len_of_A_ek \<eta>"
  and encryption: "\<And>s n' t m n m' s'. 
    \<lbrakk> (m', s') \<in> set (run_state (encryption \<eta> n t m n') s); len_of \<eta> t = length m \<rbrakk>
    \<Longrightarrow> length m' = len_of_A_e \<eta> (len_of_A_ek \<eta>) (length m) (len_of_P_nonce \<eta>)"
  and verification_key: "\<And>n s m s'. (m, s') \<in> set (run_state (verification_key \<eta> n) s) \<Longrightarrow> length m = len_of_A_vk \<eta>"
  and signature: "\<And>s n' t m n m' s'. 
    \<lbrakk> (m', s') \<in> set (run_state (signature \<eta> n t m n') s); len_of \<eta> t = length m \<rbrakk>
    \<Longrightarrow> length m' = len_of_A_sig \<eta> (len_of_A_sk \<eta>) (length m) (len_of_P_nonce \<eta>)"
  shows "len_of \<eta> t = length msg"
using \<beta>
proof(induction t arbitrary: msg s' s)
  case E thus ?case unfolding \<beta>_simps in_set_\<beta>_E using E.IH encryption pnonce
    by(auto)(rule len_of_A_, fastforce+)
next
  case SIG thus ?case unfolding in_set_\<beta>_SIG using SIG.IH signature by(auto intro: len_of_A_)
next
  case N thus ?case using pnonce by(fastforce simp add: in_set_\<beta>_N)
next
  case EK thus ?case using encryption_key by(auto simp add: in_set_\<beta>_EK)
next
  case VK thus ?case using verification_key by(auto simp add: in_set_\<beta>_VK)
next
  case DK thus ?case using pnonce
    by(clarsimp simp add: in_set_\<beta>_DK)(rule len_of_A_, fastforce+)
next
  case SK thus ?case using pnonce
    by(clarsimp simp add: in_set_\<beta>_SK)(rule len_of_A_, fastforce+)
qed(auto 4 3 simp add: in_set_\<beta> split_beta Let_def intro: len_of_A_)

end

end