theory CoSP_PKES_Symbolic imports
  CoSP
begin

section {* Computational soundness of public-key encryption and signatures *}

subsection {* Symbolic model *}

type_synonym protocol_nonce = nat

datatype (plugins del: size) nonce
  = is_PNonce: PNonce (the_PNonce: protocol_nonce)
  | ANonce bitstring -- \<open>
      We identify adversary nonces with the bitstrings they represent.
    \<close>

declare nonce.case_transfer[transfer_rule]

lemma is_PNonceE [elim!]:
  assumes "is_PNonce n"
  obtains n' where "n = PNonce n'"
using assms by(auto simp add: is_PNonce_def)

text {* We allow type flaws in terms, because the attacker can build ill-typed terms. *}

datatype t
  = E t t t
  | is_EK: EK (the_EK: t)
  | DK t
  | SIG t t t
  | VK t
  | SK t
  | PAIR t t
  | EMPTY
  | STRING bool t
  | is_N: N (the_N: nonce)
  | garbage t
  | garbageE t t
  | garbageSIG t t

declare t.case_transfer[transfer_rule]

text \<open>Zipper data type for @{typ t}\<close>
datatype t_context
   = Top
   | Ec1 (the_tc: t_context) t t
   | Ec2 t (the_tc: t_context) t
   | Ec3 t t (the_tc: t_context)
   | EKc (the_tc: t_context)
   | DKc (the_tc: t_context)
   | SIGc1 (the_tc: t_context) t t
   | SIGc2 t (the_tc: t_context) t
   | SIGc3 t t (the_tc: t_context)
   | VKc (the_tc: t_context)
   | SKc (the_tc: t_context)
   | PAIRc1 (the_tc: t_context) t
   | PAIRc2 t (the_tc: t_context)
   | STRINGc bool (the_tc: t_context)
   | garbagec (the_tc: t_context)
   | garbageEc1 (the_tc: t_context) t
   | garbageEc2 t (the_tc: t_context)
   | garbageSIGc1 (the_tc: t_context) t
   | garbageSIGc2 t (the_tc: t_context)
  where "the_tc Top = Top"

primrec up_t :: "t_context \<Rightarrow> t \<Rightarrow> t"
where
  "up_t Top t = t"
| "up_t (Ec1 c t2 t3) t = E t t2 t3"
| "up_t (Ec2 t1 c t3) t = E t1 t t3"
| "up_t (Ec3 t1 t2 c) t = E t1 t2 t"
| "up_t (EKc c) t = EK t"
| "up_t (DKc c) t = DK t"
| "up_t (SIGc1 c t2 t3) t = SIG t t2 t3"
| "up_t (SIGc2 t1 c t3) t = SIG t1 t t3"
| "up_t (SIGc3 t1 t2 c) t = SIG t1 t2 t"
| "up_t (SKc c) t = SK t"
| "up_t (VKc c) t = VK t"
| "up_t (PAIRc1 c t2) t = PAIR t t2"
| "up_t (PAIRc2 t1 c) t = PAIR t1 t"
| "up_t (STRINGc b c) t = STRING b t"
| "up_t (garbagec c) t = garbage t"
| "up_t (garbageEc1 c t2) t = garbageE t t2"
| "up_t (garbageEc2 t1 c) t = garbageE t1 t"
| "up_t (garbageSIGc1 c t2) t = garbageSIG t t2"
| "up_t (garbageSIGc2 t1 c) t = garbageSIG t1 t"

fun collapse_t :: "t_context \<Rightarrow> t \<Rightarrow> t"
where
  "collapse_t Top t = t"
| "collapse_t c t = collapse_t (the_tc c) (up_t c t)"

declare collapse_t.simps [simp del]
lemmas collapse_t_simps [simp] = collapse_t.simps[unfolded t_context.sel up_t.simps]

lemma collapse_t_unfold: "collapse_t c t = collapse_t (the_tc c) (up_t c t)"
by(cases c) simp_all

datatype constructor
   = E_C
   | EK_C
   | DK_C
   | SIG_C
   | VK_C
   | SK_C
   | PAIR_C
   | EMPTY_C
   | STRING_C bool
   -- \<open>Terms may contain nonces, but @{const N} is not a constructor in the symbolic CoSP model\<close>
   | garbage_C
   | garbageE_C
   | garbageSIG_C

inductive \<S> :: "t \<Rightarrow> bool"
where
  EMPTY: "\<S> EMPTY"
| STRING: "\<S> s \<Longrightarrow> \<S> (STRING b s)"

declare \<S>.intros[intro!]

inductive_simps S_simps [simp]:
  "\<S> (E t1 t2 t3)"
  "\<S> (EK t)"
  "\<S> (DK t)"
  "\<S> (SIG t1 t2 t3)"
  "\<S> (VK t)"
  "\<S> (SK t)"
  "\<S> (PAIR t1 t2)"
  "\<S> EMPTY"
  "\<S> (STRING b s)"
  "\<S> (N n)"
  "\<S> (garbage t)"
  "\<S> (garbageE t1 t2)"
  "\<S> (garbageSIG t1 t2)"

inductive \<T> :: "t \<Rightarrow> bool"
where
  E:  "\<T> t \<Longrightarrow> \<T> (E (EK (N n)) t (N n'))"
| EK: "\<T> (EK (N n))"
| DK: "\<T> (DK (N n))"
| SIG: "\<T> t \<Longrightarrow> \<T> (SIG (SK (N n)) t (N n'))"
| VK: "\<T> (VK (N n))"
| SK: "\<T> (SK (N n))"
| PAIR: "\<T> t1 \<Longrightarrow> \<T> t2 \<Longrightarrow> \<T> (PAIR t1 t2)"
| STRING: "\<S> t \<Longrightarrow> \<T> t"
| N: "\<T> (N n)"
| garbage: "\<T> (garbage (N n))"
| garbageE: "\<T> t \<Longrightarrow> \<T> (garbageE t (N n))"
| garbageSIG: "\<T> t \<Longrightarrow> \<T> (garbageSIG t (N n))"

declare \<T>.intros[intro!]

inductive_simps T_simps [simp]:
  "\<T> (E t1 t2 t3)"
  "\<T> (EK t)"
  "\<T> (DK t)"
  "\<T> (SIG t1 t2 t3)"
  "\<T> (VK t)"
  "\<T> (SK t)"
  "\<T> (PAIR t1 t2)"
  "\<T> EMPTY"
  "\<T> (STRING b s)"
  "\<T> (N n)"
  "\<T> (garbage t)"
  "\<T> (garbageE t1 t2)"
  "\<T> (garbageSIG t1 t2)"

inductive_cases T_cases [elim!]:
  "\<T> (E t1 t2 t3)"
  "\<T> (EK t)"
  "\<T> (DK t)"
  "\<T> (SIG t1 t2 t3)"
  "\<T> (VK t)"
  "\<T> (SK t)"
  "\<T> (PAIR t1 t2)"
  "\<T> EMPTY"
  "\<T> (STRING b s)"
  "\<T> (N n)"
  "\<T> (garbage t)"
  "\<T> (garbageE t1 t2)"
  "\<T> (garbageSIG t1 t2)"


datatype destructor
  = D
  | isenc
  | isek
  | ekof
  | verify
  | issig
  | isvk
  | vkof
  | FST
  | SND
  | unstring bool
  | equals

text {* Split the destructor application into several functions to avoid blow-up during case splitting inside fun *}

fun \<D>_D :: "t list \<rightharpoonup> t"
where
  "\<D>_D [DK t, E (EK t1) m t2] = (if t = t1 then Some m else None)"
| "\<D>_D _ = None" 

lemma \<D>_D_SomeE:
  assumes "\<D>_D x = Some t"
  obtains k r where "x = [DK k, E (EK k) t r]"
using assms by(auto elim!: \<D>_D.elims split: if_split_asm)

fun \<D>_isenc :: "t list \<rightharpoonup> t"
where
  "\<D>_isenc [E (EK t1) t2 t3] = Some (E (EK t1) t2 t3)"
| "\<D>_isenc [garbageE t1 t2] = Some (garbageE t1 t2)"
| "\<D>_isenc _ = None"

fun_cases \<D>_isenc_cases_E : "\<D>_isenc [E (EK t1) t2 t3] = Some a"
fun_cases \<D>_isenc_SomeE: "\<D>_isenc x = Some t"

fun \<D>_isek :: "t list \<rightharpoonup> t"
where
  "\<D>_isek [EK t] = Some (EK t)"
| "\<D>_isek _ = None"

fun_cases \<D>_isek_SomeE: "\<D>_isek x = Some t"

fun \<D>_ekof :: "t list \<rightharpoonup> t"
where
  "\<D>_ekof [E (EK t1) m t2] = Some (EK t1)"
| "\<D>_ekof [garbageE t1 t2] = Some t1"
| "\<D>_ekof _ = None"

fun_cases \<D>_ekof_SomeE: "\<D>_ekof x = Some t"

fun \<D>_verify :: "t list \<rightharpoonup> t"
where
  "\<D>_verify [VK t, SIG (SK t1) t2 t3] = (if t = t1 then Some t2 else None)"
| "\<D>_verify _ = None"

lemma \<D>_verify_SomeE:
  assumes "\<D>_verify x = Some t"
  obtains k r where "x = [VK k, SIG (SK k) t r]"
using assms by(auto elim!: \<D>_verify.elims split: if_split_asm)

fun \<D>_issig :: "t list \<rightharpoonup> t"
where
  "\<D>_issig [SIG (SK t1) t2 t3] = Some (SIG (SK t1) t2 t3)"
| "\<D>_issig [garbageSIG t1 t2] = Some (garbageSIG t1 t2)"
| "\<D>_issig _ = None"

fun_cases \<D>_issig_SomeE: "\<D>_issig x = Some t"

fun \<D>_isvk :: "t list \<rightharpoonup> t"
where
  "\<D>_isvk [VK t] = Some (VK t)"
| "\<D>_isvk _ = None"

fun_cases \<D>_isvk_SomeE: "\<D>_isvk x = Some t"

fun \<D>_vkof :: "t list \<rightharpoonup> t"
where
  "\<D>_vkof [SIG (SK t1) m t2] = Some (VK t1)" 
| "\<D>_vkof [garbageSIG t1 t2] = Some t1"
| "\<D>_vkof _ = None"

fun_cases \<D>_vkof_SomeE: "\<D>_vkof x = Some t"

fun \<D>_FST :: "t list \<rightharpoonup> t"
where
  "\<D>_FST [PAIR t1 t2] = Some t1"
| "\<D>_FST _ = None"

fun_cases \<D>_FST_SomeE: "\<D>_FST x = Some t"

fun \<D>_SND :: "t list \<rightharpoonup> t"
where
  "\<D>_SND [PAIR t1 t2] = Some t2"
| "\<D>_SND _ = None"

fun_cases \<D>_SND_SomeE: "\<D>_SND x = Some t"

fun \<D>_unstring :: "bool \<Rightarrow> t list \<rightharpoonup> t"
where
  "\<D>_unstring b [STRING x xs] = (if x = b then Some xs else None)"
| "\<D>_unstring _ _ = None"

lemma \<D>_unstring_SomeE:
  assumes "\<D>_unstring b x = Some t"
  obtains "x = [STRING b t]"
using assms by(auto elim!: \<D>_unstring.elims split: if_split_asm)

fun \<D>_equals :: "t list \<rightharpoonup> t"
where
  "\<D>_equals [t1, t2] = (if t1 = t2 then Some t1 else None)"
| "\<D>_equals _ = None"

lemma \<D>_equals_SomeE:
  assumes "\<D>_equals x = Some t"
  obtains "x = [t, t]"
using assms by(auto elim!: \<D>_equals.elims split: if_split_asm)

fun \<D> :: "destructor \<Rightarrow> t list \<rightharpoonup> t"
where
  "\<D> D = \<D>_D"
| "\<D> isenc = \<D>_isenc"
| "\<D> isek = \<D>_isek"
| "\<D> ekof = \<D>_ekof"
| "\<D> verify = \<D>_verify"
| "\<D> issig = \<D>_issig"
| "\<D> isvk = \<D>_isvk"
| "\<D> vkof = \<D>_vkof"
| "\<D> FST = \<D>_FST"
| "\<D> SND = \<D>_SND"
| "\<D> (unstring b) = \<D>_unstring b"
| "\<D> equals = \<D>_equals"

lemmas \<D>_cases[case_names D isenc isek ekof verify issig isvk vkof FST SND unstring equals] = \<D>.cases

lemmas \<D>s_SomeE =
  \<D>_D_SomeE 
  \<D>_isenc_SomeE
  \<D>_isek_SomeE
  \<D>_ekof_SomeE
  \<D>_verify_SomeE
  \<D>_issig_SomeE
  \<D>_isvk_SomeE
  \<D>_vkof_SomeE
  \<D>_FST_SomeE
  \<D>_SND_SomeE
  \<D>_unstring_SomeE
  \<D>_equals_SomeE

text \<open>The terms which are accessible using destructors\<close>

inductive acc_subterm1 :: "t \<Rightarrow> t \<Rightarrow> bool"
where
  E_key: "acc_subterm1 (E (EK k) t r) (EK k)"
| E_message: "acc_subterm1 (E (EK k) t r) t"
| SIG_key: "acc_subterm1 (SIG (SK k) t r) (VK k)"
| SIG_message: "acc_subterm1 (SIG (SK k) t r) t"
| PAIR1: "acc_subterm1 (PAIR t1 t2) t1"
| PAIR2: "acc_subterm1 (PAIR t1 t2) t2"
| STRING: "acc_subterm1 (STRING b t) t"
| garbageE: "acc_subterm1 (garbageE k t) k"
| garbageSIG: "acc_subterm1 (garbageSIG k t) k"

abbreviation acc_subterm :: "t \<Rightarrow> t \<Rightarrow> bool"
where "acc_subterm \<equiv> acc_subterm1\<^sup>*\<^sup>*"

inductive_simps acc_subterm1_simps [simp]:
  "acc_subterm1 (E t1 t2 t3) t"
  "acc_subterm1 (EK t1) t"
  "acc_subterm1 (DK t1) t"
  "acc_subterm1 (SIG t1 t2 t3) t"
  "acc_subterm1 (VK t1) t"
  "acc_subterm1 (SK t1) t"
  "acc_subterm1 (PAIR t1 t2) t"
  "acc_subterm1 EMPTY t"
  "acc_subterm1 (STRING b t1) t"
  "acc_subterm1 (N n) t"
  "acc_subterm1 (garbage t1) t"
  "acc_subterm1 (garbageE t1 t2) t"
  "acc_subterm1 (garbageSIG t1 t2) t"

lemma acc_subterm_EK [simp]: "acc_subterm (EK t) t' \<longleftrightarrow> t' = EK t"
by(auto elim: converse_rtranclpE)

lemma acc_subterm_E [simp]:
  "acc_subterm (E t1 t2 t3) t \<longleftrightarrow> t = E t1 t2 t3 \<or> (\<exists>k. t1 = EK k \<and> (t = EK k \<or> acc_subterm t2 t))"
by(auto 4 3 elim: converse_rtranclpE intro: converse_rtranclp_into_rtranclp)

lemma acc_subterm_DK [simp]: "acc_subterm (DK t) t' \<longleftrightarrow> t' = DK t"
by(auto elim: converse_rtranclpE)
   
lemma acc_subterm_VK [simp]: "acc_subterm (VK t) t' \<longleftrightarrow> t' = VK t"
by(auto elim: converse_rtranclpE)

lemma acc_subterm_SIG [simp]:
  "acc_subterm (SIG t1 t2 t3) t \<longleftrightarrow> t = SIG t1 t2 t3 \<or> (\<exists>k. t1 = SK k \<and> (t = VK k \<or> acc_subterm t2 t))"
by(auto 4 3 elim: converse_rtranclpE intro: converse_rtranclp_into_rtranclp)

lemma acc_subterm_SK [simp]: "acc_subterm (SK t) t' \<longleftrightarrow> t' = SK t"
by(auto elim: converse_rtranclpE)

lemma acc_subterm_PAIR [simp]:
  "acc_subterm (PAIR t1 t2) t \<longleftrightarrow> t = PAIR t1 t2 \<or> acc_subterm t1 t \<or> acc_subterm t2 t"
by(auto elim: converse_rtranclpE intro: converse_rtranclp_into_rtranclp)

lemma acc_subterm_EMPTY [simp]: "acc_subterm EMPTY t \<longleftrightarrow> t = EMPTY"
by(auto elim: converse_rtranclpE)

lemma acc_subterm_STRING [simp]: "acc_subterm (STRING b t) t' \<longleftrightarrow> t' = STRING b t \<or> acc_subterm t t'"
by(auto elim: converse_rtranclpE intro: converse_rtranclp_into_rtranclp)

lemma acc_subterm_N [simp]: "acc_subterm (N n) t \<longleftrightarrow> t = N n"
by(auto elim: converse_rtranclpE)

lemma acc_subterm_garbage [simp]: "acc_subterm (garbage t) t' \<longleftrightarrow> t' = garbage t"
by(auto elim: converse_rtranclpE)

lemma acc_subterm_garbageE [simp]:
  "acc_subterm (garbageE t1 t2) t \<longleftrightarrow> t = garbageE t1 t2 \<or> acc_subterm t1 t"
by(auto elim: converse_rtranclpE intro: converse_rtranclp_into_rtranclp)

lemma acc_subterm_garbageSIG [simp]:
  "acc_subterm (garbageSIG t1 t2) t \<longleftrightarrow> t = garbageSIG t1 t2 \<or> acc_subterm t1 t"
by(auto elim: converse_rtranclpE intro: converse_rtranclp_into_rtranclp)

inductive deduce :: "t set \<Rightarrow> t \<Rightarrow> bool" ("_/ \<turnstile> (2_)" [120, 120] 65)
  for S
where
  axiom: "m \<in> S \<Longrightarrow> S \<turnstile> m"
| nonce: "S \<turnstile> N (ANonce n)"
-- \<open>Constructors\<close>
| E: "\<lbrakk> S \<turnstile> EK (N n); S \<turnstile> t2; S \<turnstile> N n' \<rbrakk> \<Longrightarrow> S \<turnstile> E (EK (N n)) t2 (N n')"
| EK: "S \<turnstile> N n \<Longrightarrow> S \<turnstile> EK (N n)"
| DK: "S \<turnstile> N n \<Longrightarrow> S \<turnstile> DK (N n)"
| SIG: "\<lbrakk> S \<turnstile> SK (N n); S \<turnstile> t2; S \<turnstile> N n' \<rbrakk> \<Longrightarrow> S \<turnstile> SIG (SK (N n)) t2 (N n')"
| VK: "S \<turnstile> N n \<Longrightarrow> S \<turnstile> VK (N n)"
| SK: "S \<turnstile> N n \<Longrightarrow> S \<turnstile> SK (N n)"
| PAIR: "\<lbrakk> S \<turnstile> t1; S \<turnstile> t2 \<rbrakk> \<Longrightarrow> S \<turnstile> PAIR t1 t2"
| EMPTY: "S \<turnstile> EMPTY"
| STRING: "\<lbrakk> S \<turnstile> s; \<S> s \<rbrakk> \<Longrightarrow> S \<turnstile> STRING b s"
| garbage: "S \<turnstile> N n \<Longrightarrow> S \<turnstile> garbage (N n)"
| garbageE: "\<lbrakk> S \<turnstile> t; S \<turnstile> N n \<rbrakk> \<Longrightarrow> S \<turnstile> garbageE t (N n)"
| garbageSIG: "\<lbrakk> S \<turnstile> t; S \<turnstile> N n \<rbrakk> \<Longrightarrow> S \<turnstile> garbageSIG t (N n)"
-- \<open>Destructors\<close>
| D: "\<lbrakk> S \<turnstile> DK t1; S \<turnstile> E (EK t1) t2 t3 \<rbrakk> \<Longrightarrow> S \<turnstile> t2"
| verify: "\<lbrakk> S \<turnstile> VK t1; S \<turnstile> SIG (SK t1) t2 t3 \<rbrakk> \<Longrightarrow> S \<turnstile> t2"
  -- \<open>The tests @{term isenc}, @{term isek}, @{term verify}, @{term issig}, @{term isvk}, and 
      @{term equals} just return the input value if successful, so we need no deduction rules
      for them\<close>
| ekof_E: "S \<turnstile> E (EK t1) m t2 \<Longrightarrow> S \<turnstile> EK t1"
| ekof_garbage: "S \<turnstile> garbageE t1 t2 \<Longrightarrow> S \<turnstile> t1"
| vkof_E: "S \<turnstile> SIG (SK t1) m t2 \<Longrightarrow> S \<turnstile> VK t1"
| vkof_garbage: "S \<turnstile> garbageSIG t1 t2 \<Longrightarrow> S \<turnstile> t1"
| FST: "S \<turnstile> PAIR t1 t2 \<Longrightarrow> S \<turnstile> t1"
| SND: "S \<turnstile> PAIR t1 t2 \<Longrightarrow> S \<turnstile> t2"
| unstring: "S \<turnstile> STRING b t \<Longrightarrow> S \<turnstile> t"

lemma deduce_\<T>:
  assumes "S \<turnstile> t" "\<forall>t\<in>S. \<T> t"
  shows "\<T> t"
using \<open>S \<turnstile> t\<close>
proof(induction)
  case axiom with \<open>\<forall>t\<in>S. _\<close> show ?case by auto
qed auto

lemma deduce_\<D>:
  notes \<D>s_SomeE[elim!]
  assumes \<D>: "\<D> d ts = Some t"
  and deduce: "\<forall>t\<in>set ts. S \<turnstile> t"
  shows "S \<turnstile> t"
proof(cases d rule: \<D>_cases)
  { case D        with assms show ?thesis by(auto intro: deduce.D) }
  { case isenc    with assms show ?thesis by(auto) }
  { case isek     with assms show ?thesis by(auto) }
  { case ekof     with assms show ?thesis by(auto intro: deduce.ekof_E deduce.ekof_garbage) }
  { case verify   with assms show ?thesis by(auto intro: deduce.verify) }
  { case issig    with assms show ?thesis by(auto) }
  { case isvk     with assms show ?thesis by(auto) }
  { case vkof     with assms show ?thesis by(auto intro: deduce.vkof_E deduce.vkof_garbage) }
  { case FST      with assms show ?thesis by(auto intro: deduce.FST) }
  { case SND      with assms show ?thesis by(auto intro: deduce.SND) }
  { case unstring with assms show ?thesis by(auto intro: deduce.unstring) }
  { case equals   with assms show ?thesis by(auto) }
qed

fun \<C> :: "constructor \<Rightarrow> t list \<Rightarrow> t"
where
  "\<C> E_C (t1 # t2 # t3 # _) = E t1 t2 t3"
| "\<C> EK_C (t # _) = EK t"
| "\<C> DK_C (t # _) = DK t"
| "\<C> SIG_C (t1 # t2 # t3 # _) = SIG t1 t2 t3"
| "\<C> VK_C (t # _) = VK t"
| "\<C> SK_C (t # _) = SK t"
| "\<C> PAIR_C (t1 # t2 # _) = PAIR t1 t2"
| "\<C> EMPTY_C _ = EMPTY"
| "\<C> (STRING_C b) (t # _) = STRING b t"
| "\<C> garbage_C (t # _) = garbage t"
| "\<C> garbageE_C (t1 # t2 # _) = garbageE t1 t2"
| "\<C> garbageSIG_C (t1 # t2 # _) = garbageSIG t1 t2"
| "\<C> constr ts = garbage (undefined constr ts)"

fun_cases \<C>s_E:  
  "\<C> C ts = E t1 t2 t3"
  "\<C> C ts = EK t"
  "\<C> C ts = DK t"
  "\<C> C ts = SIG t1 t2 t3"
  "\<C> C ts = VK t"
  "\<C> C ts = SK t"
  "\<C> C ts = PAIR t1 t2"
  "\<C> C ts = EMPTY"
  "\<C> C ts = STRING b t"
  "\<C> C ts = garbageE t1 t2"
  "\<C> C ts = garbageSIG t1 t2"

interpretation symbolic_model \<C> N \<D> \<T> deduce .

abbreviation hybrid_step1'
  :: "('s, 'id, t, 'call, 'ret, 'more) simulator'_scheme
  \<Rightarrow> (t, 'id, constructor, nonce, destructor, 's) hybrid_state
  \<Rightarrow> ((t option \<times> (t, 'id, constructor, nonce, destructor, 's) hybrid_state) option, 'call, 'ret) gpv"
where "hybrid_step1' \<equiv> hybrid_step1 TYPE('more)"

abbreviation hybrid_step'
  :: "('s, 'id, t, 'call, 'ret, 'more) simulator'_scheme
  \<Rightarrow> (t, 'id, constructor, nonce, destructor, 's) hybrid_state
  \<Rightarrow> ((t, 'id, constructor, nonce, destructor, 's) hybrid_state option, 'call, 'ret) gpv"
where "hybrid_step' \<equiv> hybrid_step TYPE('more)"

type_synonym 'id cosp_protocol = "('id, constructor, nonce, destructor) protocol"
translations
  (type) "'id cosp_protocol" \<leftharpoondown> (type) "('id, constructor, nonce, destructor) protocol"

fun pnonces :: "t \<Rightarrow> protocol_nonce set"
where
  "pnonces (E t1 t2 t3) = pnonces t1 \<union> pnonces t2 \<union> pnonces t3"
| "pnonces (EK t) = pnonces t"
| "pnonces (DK t) = pnonces t"
| "pnonces (SIG t1 t2 t3) = pnonces t1 \<union> pnonces t2 \<union> pnonces t3"
| "pnonces (VK t) = pnonces t"
| "pnonces (SK t) = pnonces t"
| "pnonces (PAIR t1 t2) = pnonces t1 \<union> pnonces t2"
| "pnonces EMPTY = {}"
| "pnonces (STRING b t) = pnonces t"
| "pnonces (N (PNonce n)) = {n}"
| "pnonces (N (ANonce n)) = {}"
| "pnonces (garbage t) = pnonces t"
| "pnonces (garbageE t1 t2) = pnonces t1 \<union> pnonces t2"
| "pnonces (garbageSIG t1 t2) = pnonces t1 \<union> pnonces t2"

lemma finite_pnonces [simp]: "finite (pnonces t)"
by(induction t rule: pnonces.induct) auto

datatype nonce_type = Encryption | Signature | Randomness backlink
  -- \<open>Randomness may be used only once, so the typing already stores the node index in the store
  which records the computations that can use the randomness as a forward link. This ensures  
  uniqueness, because typing makes sure that randomness nonces can only be used in computation
  nodes with constructors, which show up in the list of computation nodes.
\<close>
type_synonym nonce_typing = "protocol_nonce \<Rightarrow> nonce_type option"
  -- \<open>@{const None} denotes that the node is not used for cryptographic purposes\<close>

end